#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 12 15:56:33 2023

@author: psakicki
"""

########## BEGIN IMPORT ##########
#### External modules
import copy
import matplotlib.pyplot as plt
import os 

#### geodeZYX modules
from geodezyx import conv
from geodezyx import time_series 

#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########

def export_ts_figure_pdf(fig,export_path,filename,close=False):
    """ fig can accept a int (id of a Figure)
         OR the figure Object itself """

    if type(fig) is int:
        f = plt.figure(fig)
    elif type(fig) is plt.Figure:
        f = fig

    f.set_size_inches(16.53,11.69)
    # tralala pour avoir des dates propres
    # parce que dans des cas + simples f.autofmt_xdate() suffit
    for a in f.axes[1:]:
        labels = a.get_xticklabels()
        for l in labels:
            l.set_rotation(40)
            l.set_horizontalalignment('right')
    out_path = os.path.join(export_path,filename+'.pdf')
    f.tight_layout()
    f.subplots_adjust(top=0.94)
    f.savefig(out_path,papertype='a4',format='pdf')

    if close:
        plt.close(f)

    return None

def export_ts_plot(tsin,export_path,coortype='ENU',export_type=("pdf","png"),
                   plot_B = False,close_fig_after_export=True):
    """ Very beta ...
        to be implemented : merge w/ the export_figure_pdf fct """
    # plot A avec les barres de sigma
    plt.clf()
    tsin.plot(coortype,fig=1)
    tsin.plot_discont()
    f = plt.gcf()
    f.set_size_inches(16.53,11.69)
    # tralala pour avoir des dates propres
    # parce que dans des cas + simples f.autofmt_xdate() suffit
    for a in f.axes[1:]:
        labels = a.get_xticklabels()
        for l in labels:
            l.set_rotation(40)
            l.set_horizontalalignment('right')
    f.tight_layout()
    f.subplots_adjust(top=0.92)
    for typ in export_type:
        export_file = os.path.join(export_path,tsin.stat+'a.' + typ)
        f.savefig(export_file,
                  papertype='a4',format=typ)
        log.info("plot exported in %s" , export_file)
    if close_fig_after_export:
        plt.close(f)

    # plot B avec une droite de regression
    if plot_B:
        time_series.linear_regress_ts(tsin)
        f = plt.gcf()
        #    f.set_dpi(300)
        f.set_size_inches(16.53,11.69)
        for a in f.axes:
            labels = a.get_xticklabels()
            for l in labels:
                l.set_rotation(40)
                l.set_horizontalalignment('right')
        f.tight_layout()
        f.subplots_adjust(top=0.92)
        for typ in export_type:
            export_file = os.path.join(export_path,tsin.stat+'b.' + typ)
            f.savefig(export_file,
                      papertype='a4',format=typ)
        log.info("plot exported in %s" , export_file)
        if close_fig_after_export:
            plt.close(f)
    return None


def export_ts(ts,outdir,coordtype = 'ENU',outprefix='',write_header=False):
    """
    export the timeserie

    write_header not well implemented !!!
    """

    proto_str = '{:23} ' * 14

    A,B,C,T,sA,sB,sC = ts.to_list(coordtype)
    if outprefix != '':
        outprefix = outprefix + '_'

    outfilenam = outprefix +  ts.stat + "_" +  ts.name + '.' + coordtype + '.ts.dat'
    outpath = os.path.join(outdir,outfilenam)

    filobj = open(outpath,'w+')

    if write_header:
        header = "#" + proto_str.format(*(coordtype[0],"sigma"+coordtype[0],
                                          coordtype[1],"sigma"+coordtype[1],
                                          coordtype[2],"sigma"+coordtype[2],
                                          "year","month","day",
                                          "hour","minute","seconds",
                                          "year_decimal","posix_time"))
        filobj.write(header + "\n")

    for a,b,c,t,sa,sb,sc in zip(A,B,C,T,sA,sB,sC):
        paramstr = [str(e) for e in [a,sa,b,sb,c,sc]]
        #paramstr = [e.ljust(18, '0') for e in paramstr]
        
        tt = conv.posix2dt(t)
        yr_dec_str = str(conv.dt2year_decimal(tt))
        posix_str = str(t)

        paramstr_time = list(tt.strftime("%Y %m %d %H %M %S").split())

        paramstr2 = paramstr + paramstr_time + [yr_dec_str,posix_str]
        
        outlin =  proto_str.format(*paramstr2) + "\n"
        filobj.write(outlin)
    filobj.close()

    log.info('timeserie exported in %s', outpath)
    return None


def export_ts_as_neu(tsin,outdir,outprefix,coordtype = 'ENU'):
    """
    export to a HECTOR .neu compatible format

    outfile will be writed in
    /outdir/outprefixSTAT.neu
    
    NB: The XYZ mode is quite dirty (191001)
    """
    if not hasattr(tsin[0],'X'):
        log.warning('no XYZ in ts')
        noXYZ = True
    else:
        noXYZ = False

    tswork = copy.deepcopy(tsin)
    #if coordtype == 'XYZ':
    #    mp = tswork.mean_posi()
    #    tswork.ENUcalc(mp)
    outpath = outdir +'/' + outprefix + tswork.stat + '.neu'
    outfile = open(outpath,'w+')
    E,N,U,T,sE,sN,sU = tswork.to_list(coordtype)
    first_pt = tswork.pts[0]
    if noXYZ:
        first_pt.X = 0.
        first_pt.Y = 0.
        first_pt.Z = 0.
        first_pt.L = 0.
        first_pt.H = 0.
        first_pt.F = 0.


    e0,n0,u0,t0 = list(zip(E,N,U,T))[0]
    # write the header
    outfile.write('# Site : {} \n'.format(tswork.stat))
    if 'calc_center' in list(tswork.anex.keys()):
        outfile.write('# Analysis Centre: {} \n'.format(tswork.anex['calc_center']))
    else:
        outfile.write('# Analysis Centre: N/A \n')

    outfile.write('# Solution code: GINS_PS \n')
    outfile.write('# Datum: ITRF2008\n')
    outfile.write('#\n')
    outfile.write('# Reference epoch: {}\n'.format(conv.dt2year_decimal(first_pt.Tdt)))
    outfile.write('# X : {}\n'.format(first_pt.X))
    outfile.write('# Y : {}\n'.format(first_pt.Y))
    outfile.write('# Z : {}\n'.format(first_pt.Z))
    outfile.write('#\n')
    outfile.write('# Longitude : {}\n'.format(first_pt.L))
    outfile.write('# Latitude  : {}\n'.format(first_pt.F))
    outfile.write('# Height    : {}\n'.format(first_pt.H))
    outfile.write('#\n')
    if coordtype == "ENU":
        outfile.write('# Components : ' + "NEU" + "\n")
    elif coordtype == "XYZ":
        outfile.write('# Components : ' + "YXZ" + "\n")
        outfile.write('# Cartesian components are undirect to maintain consistency with NEU\n')                
    outfile.write('#\n')
    if tswork.bool_discont:
        outfile.write('# type_of_offset : from discontinuties got from a station.info\n')
        outfile.write('#\n')
        for disc in sorted(tswork.discont):
            outfile.write('# offset {} 7\n'.format(conv.dt2year_decimal(disc)))
        outfile.write('#\n')
    # write the data
    for e,n,u,t,se,sn,su in zip(E,N,U,T,sE,sN,sU):
        t = conv.dt2year_decimal(conv.posix2dt(t))
        if coordtype == "ENU":        
            outfile.write('{:.5f}   {:+.6f}    {:+.6f}    {:+.6f} {:+.6f} {:+.6f} {:+.6f}\n'.format(t,n-n0,e-e0,u-u0,se,sn,su))
        elif coordtype == "XYZ":
            outfile.write('{:.5f}   {:+.6f}    {:+.6f}    {:+.6f} {:+.6f} {:+.6f} {:+.6f}\n'.format(t,n-n0,e-e0,u-u0,se,sn,su))

    log.info('timeserie exported in %s',outpath)
    return None


def export_ts_as_hector_enu(tsin,outdir,outprefix,coordtype = 'ENU'):
    """
    export to a HECTOR .enu (and not .neu !) compatible format
    This format is simpler : just gives MJD E N U

    This format is necessary to force a sampling period.

    outfile will be writed in
    /outdir/outprefixSTAT.enu
    """

    log.warning("NOT IMPLEMENTED YET !")

    return None

def export_ts_as_midas_tenu(tsin,outdir,outprefix,coordtype = 'ENU',
                            export_step=True):
    """
    export to a MIDAS .tneu compatible format

    outfile will be writed in
    /outdir/outprefixSTAT.tneu

    if export_step == True:
    export a step file as
    /outdir/outprefixSTAT.step
    """
    if not hasattr(tsin[0],'X'):
        log.warning('no XYZ in ts')
        noXYZ = True
    else:
        noXYZ = False

    tswork = copy.deepcopy(tsin)
    stat = tswork.stat
    if coordtype == 'XYZ':
        mp = tswork.mean_posi()
        tswork.ENUcalc(mp)
    outpath = outdir +'/' + outprefix + tswork.stat + '.tenu'
    outfile = open(outpath,'w+')
    E,N,U,T,sE,sN,sU = tswork.to_list('ENU')
    first_pt = tswork.pts[0]
    if noXYZ:
        first_pt.X = 0.
        first_pt.Y = 0.
        first_pt.Z = 0.
        first_pt.L = 0.
        first_pt.H = 0.
        first_pt.F = 0.

    e0,n0,u0,t0 = list(zip(E,N,U,T))[0]

    for e,n,u,t in zip(E,N,U,T):
        t = conv.dt2year_decimal(conv.posix2dt(t))
        #outfile.write('{} {:.5f} {:+.6f} {:+.6f} {:+.6f} \n'.format(stat,t,n-n0,e-e0,u-u0))
        outfile.write('{} {:.5f} {:+.6f} {:+.6f} {:+.6f} \n'.format(stat,t,e-e0,n-n0,u-u0))

    log.info('timeserie exported in %s',outpath)

    if export_step and tswork.bool_discont:
        outpath_step = outdir +'/' + outprefix + tswork.stat + '.step'
        outfile_step = open(outpath_step,'w+')
        for d in tswork.discont:
            d = conv.dt2year_decimal(d)
            line = tswork.stat + " " + str(d) + "\n"
            outfile_step.write(line)

            log.info('timeserie discont. (steps) exported in %s',outpath_step)

    return None


###################################################################
def export_ts_as_pbo_pos(tsin, outdir, outprefix='' ,
                         force=None, force_1st_pt_as_ref=True,
                         verbose=False):
###################################################################
    """
    Write a time series in GAMIT/GLOBK PBO pos format

    :param idir: output directory
    :param outprefix: if not blank then the output pos file will be CODE_add_key.pos, CODE.pos otherwise.
    :param force: set force to 'data' or 'data_xyz' to force pos to be written from .data or .data_xyz

    :note1:default behaviour (force = None)
        if data and data_xyz are not None, then print them independently
        if there are data only, then uses X0,Y0,Z0 to write data_xyz
        if there are data_xyz only, recreate data and write it
    
    """

    # force option
    
    # if force is not None:
    #     if force not in ['data','data_xyz']:
    #         print('!!! ERROR: force keyword must be either data or data_xyz')
    #         return(self)
    
    #     if force == 'data':
    #         self.data_xyz = None

    #     if force == 'data_xyz':
    #         self.data = None

    # # cdata
    
    # if force != 'data_xyz':
    #     if not self.cdata(data=True):
    #         print('! Can not write .pos file. Problem with .data attribute.')
    #         self.cdata(data=True,verbose=True)
    #         return(self)


    ###############################################
    def __print_header_pos_file(f_pos,code,
                                first_epoch,last_epoch,
                                release_date,XYZ_ref,
                                geo_ref,verbose=False):
        
        f_pos.write("PBO Station Position Time Series. Reference Frame : Unknown\n")
        f_pos.write("Format Version: 1.1.0\n")
        f_pos.write("4-character ID: %s\n" % code)
        f_pos.write("Station name  : %s        \n" % code)
        f_pos.write("First Epoch   : %s\n" % first_epoch)
        f_pos.write("Last Epoch    : %s\n" % last_epoch)
        f_pos.write("Release Date  : %s\n" % release_date)
        f_pos.write("XYZ Reference position :   %s\n" % XYZ_ref)
        f_pos.write("NEU Reference position :   %s (Unknown/WGS84)\n" % geo_ref)
        f_pos.write("Start Field Description\n")
        f_pos.write("YYYYMMDD      Year, month, day for the given position epoch\n")
        f_pos.write("HHMMSS        Hour, minute, second for the given position epoch\n")
        f_pos.write("JJJJJ.JJJJJ   Modified Julian day for the given position epoch\n")
        f_pos.write("X             X coordinate, Specified Reference Frame, meters\n")
        f_pos.write("Y             Y coordinate, Specified Reference Frame, meters\n")
        f_pos.write("Z             Z coordinate, Specified Reference Frame, meters\n")
        f_pos.write("Sx            Standard deviation of the X position, meters\n")
        f_pos.write("Sy            Standard deviation of the Y position, meters\n")
        f_pos.write("Sz            Standard deviation of the Z position, meters\n")
        f_pos.write("Rxy           Correlation of the X and Y position\n")
        f_pos.write("Rxz           Correlation of the X and Z position\n")
        f_pos.write("Ryz           Correlation of the Y and Z position\n")
        f_pos.write("Nlat          North latitude, WGS-84 ellipsoid, decimal degrees\n")
        f_pos.write("Elong         East longitude, WGS-84 ellipsoid, decimal degrees\n")
        f_pos.write("Height (Up)   Height relative to WGS-84 ellipsoid, m\n")
        f_pos.write("dN            Difference in North component from NEU reference position, meters\n")
        f_pos.write("dE            Difference in East component from NEU reference position, meters\n")
        f_pos.write("du            Difference in vertical component from NEU reference position, meters\n")
        f_pos.write("Sn            Standard deviation of dN, meters\n")
        f_pos.write("Se            Standard deviation of dE, meters\n")
        f_pos.write("Su            Standard deviation of dU, meters\n")
        f_pos.write("Rne           Correlation of dN and dE\n")
        f_pos.write("Rnu           Correlation of dN and dU\n")
        f_pos.write("Reu           Correlation of dEand dU\n")
        f_pos.write("Soln          corresponding to products  generated with rapid or final orbit products, in supplemental processing, campaign data processing or reprocessing\n")
        f_pos.write("End Field Description\n")
        f_pos.write("*YYYYMMDD HHMMSS JJJJJ.JJJJ         X             Y             Z            Sx        Sy       Sz     Rxy   Rxz    Ryz            NLat         Elong         Height         dN        dE        dU         Sn       Se       Su      Rne    Rnu    Reu  Soln\n")

    ###############################################    

    def __format_data(tsinin):
                
        lformatted_out = []
        
        for pt in tsinin.pts:
            YYYYMMDD = int(conv.dt2str(pt.Tdt,"%Y%m%d"))
            HHMMSS = int(conv.dt2str(pt.Tdt,"%H%M%S"))
            MJD = conv.dt2MJD(pt.Tdt)
            X, Y, Z = pt.X, pt.Y, pt.Z
            Sx, Sy, Sz = pt.sX, pt.sY, pt.sZ  
            Rxy,Rxz,Ryz = 0.,0.,0.
            Elong,NLat,Height = pt.F,pt.L,pt.H
            dN,dE,dU = pt.E,pt.N,pt.U
            Sn,Se,Su = pt.sE,pt.sN,pt.sU
            Rne,Rnu,Reu = 0.,0.,0.
            Soln = "XXXXXXXXXXXXXX"
            
            l = [YYYYMMDD,HHMMSS,MJD,X,Y,Z,Sx,Sy,Sz,Rxy,Rxz,Ryz,Elong,NLat,
                 Height,dN,dE,dU,Sn,Se,Su,Rne,Rnu,Reu,Soln]
            
            lformatted_out.append(l)
            
        return lformatted_out
            
    
    ###############################################
    def __print_content_pos_file(f_pos,lformatted):

        for l in lformatted:
                        
            #l = [YYYYMMDD,HHMMSS,MJD,X,Y,Z,Sx,Sy,Sz,Rxy,Rxz,Ryz,Elong,NLat,Height,dN,dE,dU,Sn,Se,Su,Rne,Rnu,Reu,Soln]
            line_proto = " {:8d} {:6d} {:10.4f} {:14.5f} {:14.5f} {:14.5f} {:8.5f} {:8.5f} {:8.5f} {:6.3f} {:6.3f} {:6.3f}     {:14.10f}  {:14.10f} {:10.5f}  {:10.5f}{:10.5f}{:10.5f}   {:8.5f} {:8.5f} {:8.5f} {:6.3f} {:6.3f} {:6.3f} {:s}\n"
            
            f_pos.write(line_proto.format(*l))

    ###############################################


    tswork = copy.deepcopy(tsin)
    
    if not tswork.boolENU:
        tswork.ENUcalc_from_first_posi()
    else:
        first_pt = tswork.pts[0]
        if first_pt.E + first_pt.N + first_pt.U != 0. and force_1st_pt_as_ref:
            tswork.ENUcalc_from_first_posi()


    # output directory
    if not os.path.isdir(outdir):
        try:
            os.mkdir(outdir)
            print("-- Creating directory ",outdir)
        except:
            print("!!! Error : Could not create directory ",outdir)
    

        
    outpath = outdir +'/' +  tswork.stat + outprefix + '.pos'
        
    # open pos file
    f_pos=open(outpath,'w+')
    
    import datetime
    
    current_date=datetime.datetime.now()
    release_date= conv.dt2str(current_date) #("%d%02d%02d %02d%02d%02d"% (current_date.year,current_date.month,current_date.day,current_date.hour,current_date.minute,current_date.second))
    
    refpt = tswork.refENU
    
    XYZ_ref=(" %14.5lf %14.5lf %14.5lf (unknown)" % (refpt.X,refpt.Y,refpt.Z))
    geo_ref=(" %16.10lf %16.10lf %16.10lf (unknown)" % (refpt.F,refpt.L,refpt.H))
    
    __print_header_pos_file(f_pos,
                            tswork.stat,
                            tswork.startdate(),
                            tswork.enddate(),
                            release_date,
                            XYZ_ref,
                            geo_ref)
    
    
    wpos = __format_data(tswork)

    __print_content_pos_file(f_pos,wpos)
    
    f_pos.close()
    
    return





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.files_rw contains reading functions to 
import GNSS logsheets/sitelogs.

it can be imported directly with:
from geodezyx import files_rw

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


########## BEGIN IMPORT ##########
#### External modules
import copy
import datetime as dt
import dateutil
import glob
import os 
import pandas as pd
import re 

#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils

#### Import the logger
import logging
log = logging.getLogger(__name__)


##########  END IMPORT  ##########

#  _                     _               _                            _
# | |                   | |             | |                          | |
# | |     ___   __ _ ___| |__   ___  ___| |_ ___   _ __ ___  __ _  __| | ___ _ __
# | |    / _ \ / _` / __| '_ \ / _ \/ _ \ __/ __| | '__/ _ \/ _` |/ _` |/ _ \ '__|
# | |___| (_) | (_| \__ \ | | |  __/  __/ |_\__ \ | | |  __/ (_| | (_| |  __/ |
# |______\___/ \__, |___/_| |_|\___|\___|\__|___/ |_|  \___|\__,_|\__,_|\___|_|
#               __/ |
#              |___/
#



# ===== functions =====

def read_blocks_logsheet(input_file, block_id):
    """
    Read a Logsheet Block

    Parameters
    ----------
    input_file : str
        Logsheet path.
    block_id : int
        the ID of the wished block.
        1.   Site Identification of the GNSS Monument
        2.   Site Location Information 
        3.   GNSS Receiver Information
        4.   GNSS Antenna Information
        5.   Surveyed Local Ties
        6.   Frequency Standard
        7.   Collocation Information
        8.   Meteorological Instrumentation
        etc. etc. ...

    Returns
    -------
    ObjList : List of Block Objects
         List of Block Objects.
         Warning: if the Block contains only one sub-block
         Then the output is a 1-item List
    """
    proto_objects = [None,Site(),Location(),Receiver(),Antenna()]

    blkstr = str(block_id) + '.'
    blkstrnxt = str(block_id+1) + '.'
    blkstrX = str(block_id) + '.x'

    ObjList = []
    inblock = False
    

    for line in open(input_file,'rb'):
        
        ### manage unknown caracters
        line = str(line.decode('utf-8',errors='ignore'))


        if line == '\n':
            continue

        if line.startswith(blkstrnxt) or line.startswith(blkstrX):
            inblock = False
            break

        if line.startswith(blkstr) and (not 'Information' in line or block_id == 2):
            Obj = copy.copy(proto_objects[block_id])
            ObjList.append(Obj)
            inblock = True

        if inblock:
            if 'Approximate Position' in line:
                continue
            prop = line[4:30].strip().replace(' ', '_').replace('/','_').replace('(','').replace(')','').replace('.','').replace(',','_')
            if 'Marker->ARP Up Ecc' in prop:
                prop = 'Up_Ecc'
            elif 'Marker->ARP North Ecc' in prop:
                prop = 'North_Ecc'
            elif 'Marker->ARP East Ecc' in prop:
                prop = 'East_Ecc'
            elif 'Latitude (N is +)' in prop:
                prop = 'Latitude'
            elif 'Longitude (E is +)' in prop:
                prop = 'Longitude'
            elif 'Elevation (m,ellips.)' in prop:
                prop = 'Elevation'
            data = line[31:].strip()
            try:
                setattr(Obj,prop,float(data))
            except:
                setattr(Obj,prop,data)

    return ObjList

def mono_logsheet_read(logsheet_path,return_lists = False):
    """
    Read a sigle LogSheet File to Logsheet block objects

    Parameters
    ----------
    logsheet_path : str
        Logsheet path.
    return_lists : Bool, optional
        The default is False.
        if True, this legacy mode returns each Logsheet Block object in a list, 
        so each of them can be managed immediatly by
        write_station_info_from_datalists function.
        Useful only for such write_<...> functions.

    Returns
    -------
    period_lis : List of Period Block Object
        Period description in Period Block Object.
    sit : Site Block Object
        Site description in Site Block Object.
    loc : Location Block Object
        Location description in Site Block Object.
    """
    
    sit_lis = read_blocks_logsheet(logsheet_path,1)
    loc_lis = read_blocks_logsheet(logsheet_path,2)
    rec_lis = read_blocks_logsheet(logsheet_path,3)
    ant_lis = read_blocks_logsheet(logsheet_path,4)

    ant_install = [e.Date_Installed for e in ant_lis]
    rec_install = [e.Date_Installed for e in rec_lis]

    # merging all install date
    all_install_date = sorted(list(set(ant_install + rec_install )))

    # for each install date, find the Rec/Ant couple
    date_ant_rec_couple_lis = []
    for d in all_install_date:
        # exclude extremal date
        if d > dt.datetime(2098,1,1,tzinfo=dateutil.tz.tzutc()):
            continue

        potential_ant = []
        potential_rec = []

        for a in ant_lis:
            if a.Date_Installed <= d < a.Date_Removed:
                potential_ant.append(a)
        for r in rec_lis:
            if r.Date_Installed <= d < r.Date_Removed:
                potential_rec.append(r)

        site_code = sit_lis[0].Four_Character_ID
        if len(potential_ant) == 0:
            log.warning('%s: missing Antenna, %s, skip ...',site_code,d)
            continue
        if len(potential_rec) == 0:
            log.warning('%s: missing Receiver, %s, skip ...',site_code,d)
            continue
        if len(potential_ant) != 1:
            log.warning('s%: several Antennas found, %s',site_code,d)
        if len(potential_rec) != 1:
            log.warning('s%: several Receivers found, %s',site_code,d)

        date_ant_rec_couple_lis.append((d , potential_ant[0] , potential_rec[0]))

    if len(date_ant_rec_couple_lis) != len(set(date_ant_rec_couple_lis)):
        log.debug('bug 2')

    # construction of a period
    period_lis = []
    for i in range(len(date_ant_rec_couple_lis)):
        d1 = date_ant_rec_couple_lis[i][0]
        if i+1 == len(date_ant_rec_couple_lis):
            d2 = dt.datetime(2099,1,1,tzinfo=dateutil.tz.tzutc())
        else:
            d2 = date_ant_rec_couple_lis[i+1][0]
        a = date_ant_rec_couple_lis[i][1]
        r = date_ant_rec_couple_lis[i][2]

        period_lis.append((d1 , d2 , a , r))

    sit = sit_lis[0]
    loc = loc_lis[0]


    if return_lists:
        return [period_lis] , [sit] , [loc]
    else:
        return period_lis , sit , loc


def multi_logsheet_read(pathin,wildcardin='*log',return_dico=True,
                        output_mode='classic'):

    """
    Read multiple logsheets

    Parameters
    ----------
    pathin : list or str.
        If list: list of the Logsheet paths.
        If str: path of the directory containing the Logsheets.
    wildcardin : str, optional
        a wildcard describing the logsheet pattern.
        used only for pathin in a string path.
        The default is '\*log'.
    return_dico : bool, optional
        If False, returns period_lis_lis , stat_lis , loc_lis
        the False mode is useful for station.info generation
        with write_station_info_from_datalists
        The default is True.
    output_mode : str, optional
        Defines the output mode if a dictionary is asked.
        (It is not relevant if list are asked)
        'classic': returns in the period_lis , stat , loc
        'pretty': returns the period_lis, stat, loc but with 
        period_lis as a DataFrame
        'legacy': returns in the dico period_lis_lis , stat_lis , loc_lis 
        (should not be used anymore)
        The default is 'classic'.
        
    Returns
    -------
    stations_dico : dict
        a dictionnary like stations_dico['STAT'] = (period_lis, station, loc).

    """
    
    if utils.is_iterable(pathin):
        logsheet_list = pathin
    else:
        fullpath = os.path.join(pathin,wildcardin)
        logsheet_list = sorted(glob.glob(fullpath))
        
        if not logsheet_list:
            log.error("no logsheets found, exiting ...")
            log.error(fullpath)
            return None
        
    log.info("logsheet_list: ")
    log.info(logsheet_list)

    period_lis_lis = []
    stat_lis       = []
    loc_lis        = []

    stations_dico = dict()

    for ls in logsheet_list:
        try:
            p,s,l = mono_logsheet_read(ls,return_lists=True)
        except:
            log.warning("%s skipped for unknown reason ...",ls)
            log.warning("       logsheet must be checked")
            continue

        period_lis_lis = period_lis_lis + p
        stat_lis       = stat_lis       + s
        loc_lis        = loc_lis        + l
        
        if output_mode == "classic":
            stations_dico[s[0].Four_Character_ID] = (p[0],s[0],l[0])
        elif output_mode == "pretty":
            p_DF = pd.DataFrame(p[0])
            p_DF.columns = ('start','end','ant','rec')
            stations_dico[s[0].Four_Character_ID] = (p_DF,s[0],l[0])
        elif output_mode == "legacy":
            stations_dico[s[0].Four_Character_ID] = (p,s,l)
        else:
            log.error('check output_mode value')
            raise Exception   


    if not return_dico:
        return period_lis_lis , stat_lis , loc_lis
    else:
        return stations_dico
    
# ============================= OBJECTS ======================================

class Event(object):
    def __init__(self):
        self.__Date_Installed            =  dt.datetime(1980,1,1)
        self.__Date_Removed              =  dt.datetime(2099,1,1)

    @property
    def Date_Installed(self):
        return self.__Date_Installed

    @Date_Installed.setter
    def Date_Installed(self,indate):
        if isinstance(indate,dt.datetime):
            self.__Date_Installed = indate
        elif 'CCYY-MM-DD' in indate:
            self.__Date_Installed = dt.datetime(1980,1,1,
                                                tzinfo=dateutil.tz.tzutc())
        elif 'XXXX' in indate:
            self.__Date_Installed = dt.datetime(1980,1,1,
                                                tzinfo=dateutil.tz.tzutc())
        else:
            self.__Date_Installed = dateutil.parser.parse(indate).replace(hour=0,
            minute=0, second=0, microsecond=0,tzinfo=dateutil.tz.tzutc())

    @property
    def Date_Removed(self):
        return self.__Date_Removed

    @Date_Removed.setter
    def Date_Removed(self,indate):
        if isinstance(indate,dt.datetime):
            self.__Date_Removed = indate
        elif 'CCYY-MM-DD' in indate:
            self.__Date_Removed = dt.datetime(2099,1,1,
                                              tzinfo=dateutil.tz.tzutc())
        elif 'XXXX' in indate:
            self.__Date_Removed = dt.datetime(2099,1,1,
                                              tzinfo=dateutil.tz.tzutc())
        else:
            self.__Date_Removed = dateutil.parser.parse(indate).replace(hour=0,
            minute=0, second=0, microsecond=0,tzinfo=dateutil.tz.tzutc())

class Receiver(Event):
    def __init__(self):
        Event.__init__(self)
        self.Receiver_Type             =  'XXXX'
        self.Satellite_System          =  'XXXX'
        self.Serial_Number             =  'XXXX'
        self.Firmware_Version          =  'XXXX'
        self.Elevation_Cutoff_Setting  =  'XXXX'
        self.Temperature_Stabiliz      =  'XXXX'
        self.Additional_Information    =  'XXXX'
#        self.Date_Installed            = 'XXXX'
#        self.Date_Removed              = 'XXXX'

    def __repr__(self):
        return "Receiver Type  :{},\nDate Installed:{},\nDate Removed  :{}".format(self.Receiver_Type,
                                 self.Date_Installed,
                                 self.Date_Removed)


    def FirmwareSmart(self):
        aaa = str(self.Firmware_Version).split()
        for a in aaa:
            try:
                return float(a)
            except:
                continue
        return 0.


class Antenna(Event):
    def __init__(self):
        Event.__init__(self)
        self.Antenna_Type             =  'XXXX'
        self.Serial_Number            =  'XXXX'
        self.Antenna_Reference_Point  =  'XXXX'
        self.Up_Ecc                   =  0.
        self.North_Ecc                =  0.
        self.East_Ecc                 =  0.
        self.Alignment_from_True_N    =  0.
        self.__Antenna_Radome_Type    =  'XXXX'
        self.Radome_Serial_Number     =  'XXXX'
        self.Antenna_Cable_Type       =  'XXXX'
        self.Antenna_Cable_Length     =  'XXXX'
        self.Additional_Information   =  'XXXX'

    @property
    def Antenna_Radome_Type(self):
        return self.__Antenna_Radome_Type

    @Antenna_Radome_Type.setter
    def Antenna_Radome_Type(self,inradome):
        if inradome == '':
            self.__Antenna_Radome_Type = 'NONE'
        else:
            self.__Antenna_Radome_Type = inradome[0:4].upper()

    def __repr__(self):
        return "Antenna Type  :{},\nDate Installed:{},\nDate Removed  :{}".format(self.Antenna_Type,self.Date_Installed,self.Date_Removed)

    def AntTypSmart(self):
        #Elimination of the Radome type
        if len(self.Antenna_Type.split()) > 1:
            return self.Antenna_Type.split()[0]
        else:
            return self.Antenna_Type

    def ARPSmart(self):
        #return 'DH'+self.Antenna_Reference_Point
        # see http://rses.anu.edu.au/geodynamics/gps/papers/gamit/apdx2_ps.pdf
        #return 'DHPAB'
        return 'DHARP'


class Site(object):
    def __init__(self):
        self.Site_Name               =  'XXXX'
        self.Four_Character_ID       =  'XXXX'
        self.Monument_Inscription    =  'XXXX'
        self.__IERS_DOMES_Number     =  'XXXX'
        self.CDP_Number              =  'XXXX'
        self.Monument_Description    =  'XXXX'
        self.Height_of_the_Monument  =  'XXXX'
        self.Monument_Foundation     =  'XXXX'
        self.Foundation_Depth        =  'XXXX'
        self.Marker_Description      =  'XXXX'
        self.Date_Installed          =  'XXXX'
        self.Geologic_Characteristic =  'XXXX'
        self.Bedrock_Type            =  'XXXX'
        self.Bedrock_Condition       =  'XXXX'
        self.Fracture_Spacing        =  'XXXX'
        self.Fault_zones_nearby      =  'XXXX'
        self.Distance_activity       =  'XXXX'
        self.Additional_Information  =  'XXXX'

    @property
    def IERS_DOMES_Number(self):
        return  self.__IERS_DOMES_Number
    @IERS_DOMES_Number.setter
    def IERS_DOMES_Number(self,indomes):
        #if indomes == '':
        #    self.__IERS_DOMES_Number = '99999M001'
        #else:
        #    self.__IERS_DOMES_Number = indomes

        if re.search('[0-9]{4}M[0-9]{3}', indomes.strip()):
            self.__IERS_DOMES_Number = indomes.strip()
        else:
            self.__IERS_DOMES_Number = '99999M001'

    def __repr__(self):
        return "Site Name   :{},\nCode        :{},\nInstallation:{}".format(self.Site_Name,
                                 self.Four_Character_ID,
                                 self.Date_Installed)


class Location(object):
    def __init__(self):
        self.City_or_Town            = 'XXXX'
        self.State_or_Province       = 'XXXX'
        self.Country                 = 'XXXX'
        self.Tectonic_Plate          = 'XXXX'
        self.X_coordinate_m          = 0.
        self.Y_coordinate_m          = 0.
        self.Z_coordinate_m          = 0.
        self.Latitude                = 0.
        self.Longitude               = 0.
        self.Elevation_m_ellips      = 0.
        self.Additional_Information  = 'XXXX'

        self.X_velocity              = 0.
        self.Y_velocity              = 0.
        self.Z_velocity              = 0.

        self.X_coordinate_sigma      = 0.008
        self.Y_coordinate_sigma      = 0.008
        self.Z_coordinate_sigma      = 0.008

        self.X_velocity_sigma        = 0.008
        self.Y_velocity_sigma        = 0.008
        self.Z_velocity_sigma        = 0.008

        self.Reference_epoch         = dt.datetime(2005,1,1)

    def __repr__(self):
        return "X:{},Y:{},Z:{},\nRef. epoc.:{}".format(self.X_coordinate_m,
                                                     self.Y_coordinate_m,
                                                     self.Z_coordinate_m,
                                                     self.Reference_epoch)

    def export_as_string(self):
        Str_list = []
        Str_list.append('City or Town             : {:}'.format(self.City_or_Town))
        Str_list.append('State or Province        : {:}'.format(self.State_or_Province))
        Str_list.append('Country                  : {:}'.format(self.Country))
        Str_list.append('Tectonic Plate           : {:}'.format(self.Tectonic_Plate))
        Str_list.append('Approximate Position (ITRF)')

        X = self.X_coordinate_m
        Y = self.Y_coordinate_m
        Z = self.Z_coordinate_m

        lat , lon , h = conv.XYZ2GEO(X,Y,Z)
        lat_deg , lat_min , lat_sec = conv.deg2deg_dec2dms(lat)
        lon_deg , lon_min , lon_sec = conv.deg2deg_dec2dms(lon)

        Str_list.append('X coordinate (m)       : {:}'.format(X))
        Str_list.append('Y coordinate (m)       : {:}'.format(Y))
        Str_list.append('Z coordinate (m)       : {:}'.format(Z))

        Str_list.append('Latitude (N is +)      : {:+3d}{:2d}{:5.2d}'.format((lat_deg,lat_min,lat_sec)))
        Str_list.append('Longitude (E is +)     : {:+3d}{:2d}{:5.2d}'.format((lat_deg,lat_min,lat_sec)))
        Str_list.append('Elevation (m,ellips.)  : {:7.1f}'.format(h))
        Str_list.append('Additional Information   : N/A')

        out_str = "\n".join(Str_list)

        return out_str



# -*- coding: utf-8 -*-
"""
Created on Fri Aug  2 17:38:41 2019

@author: psakicki
"""

########## BEGIN IMPORT ##########
#### External modules
import copy
import datetime as dt
import matplotlib.pyplot as plt
import numpy as np
import os 
import scipy
import itertools

#### geodeZYX modules
from geodezyx import conv
from geodezyx import stats
from geodezyx import utils
from geodezyx import time_series
from geodezyx import reffram

#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########



def print4compar(dA,dB,dC,dD,coortype):

    if coortype == 'ENU':
        Astr,Bstr,Cstr = 'E','N','U'
    elif coortype == 'XYZ':
        Astr,Bstr,Cstr = 'X','Y','Z'
    else:
        Astr,Bstr,Cstr = 'A','B','C'

    log.info("moyenne aritm & RMS et std composante %s", Astr)
    log.info(str(np.nanmean(dA)))
    log.info(stats.RMSmean(dA))
    log.info(str(np.nanstd(dA)))
    log.info('')

    log.info("moyenne aritm & RMS et std composante %s", Bstr)
    log.info(str(np.nanmean(dB)))
    log.info(stats.RMSmean(dB))
    log.info(str(np.nanstd(dB)))
    log.info('')

    log.info("moyenne aritm & RMS et std composante %s", Cstr)
    log.info(str(np.nanmean(dC)))
    log.info(stats.RMSmean(dC))
    log.info(str(np.nanstd(dC)))
    log.info('')

    log.info("moyenne aritm & RMS et std composante D")
    log.info(str(np.nanmean(dD)))
    log.info(stats.RMSmean(dD))
    log.info(str(np.nanstd(dD)))
    log.info('')

    log.info("RMS3D : sqrt((RMS_{}**2 + RMS_{}**2 + RMS_{}**2)/3 ) ".format(Astr,Bstr,Cstr))
    log.info(stats.RMSmean([stats.RMSmean(dA),stats.RMSmean(dB),stats.RMSmean(dC)]))
    log.info("RMS2D : uniquement sur les 2 composantes plani")
    log.info(stats.RMSmean([stats.RMSmean(dA),stats.RMSmean(dB)]))
    log.info('')

def print4compar_tabular(dicolist,split=0,print_2D3D_if_any=True):

    dAlist,dBlist,dClist,dDlist,dDblist,statlist = [],[],[],[],[],[]
    for dic in dicolist:
        if "dD2D" in list(dic.keys()):
            D2Dn3D = True
        else:
            D2Dn3D = False

        if not print_2D3D_if_any:
            D2Dn3D = False

        coortype = dic['coortype']
        Dtype = dic['Dtype']

        dAlist.append(dic['dA'])
        dBlist.append(dic['dB'])
        dClist.append(dic['dC'])

        if not D2Dn3D:
            dDlist.append(dic['dD'])
        else:
            dDlist.append(dic['dD2D'])
            dDblist.append(dic['dD3D'])

        statlist.append(dic['name'])

    if coortype == 'ENU':
        Astr,Bstr,Cstr = 'E','N','U'
    elif coortype == 'XYZ':
        Astr,Bstr,Cstr = 'X','Y','Z'
    else:
        Astr,Bstr,Cstr = 'A','B','C'

    datacol = ("Arithm. mean", "RMS mean", "std. dev.")
    datacol = ("Arit.", "RMS", utils.greek_alphabet()[17])

    headerline = ['Experiment']

    if not D2Dn3D:
        headervartup = (Astr,Bstr,Cstr,'D ' + Dtype)
    else:
        headervartup = (Astr,Bstr,Cstr,'D 2D', 'D 3D')

    for abc in headervartup:
        for dc in datacol:
            headerline.append(' '.join((dc,abc)))

#    headerline.append("RMS3D : sqrt((RMS_{}**2 + RMS_{}**2 + RMS_{}**2)/3 )".format(Astr,Bstr,Cstr))
#    headerline.append("RMS2D : uniquement sur les 2 composantes plani")

    headerline.append("RMS3D")
    headerline.append("RMS2D")

    LINES_STK = [headerline]

    if not D2Dn3D:
        dDblist = [None] * len(dDlist)

    for (dA,dB,dC,dD,dDb,stat) in zip(dAlist,dBlist,dClist,dDlist,dDblist,statlist):

        if np.isclose(np.sum(dA) + np.sum(dB) + np.sum(dC),0):
            continue

        line = []
        line.append(stat)

        line.append(np.nanmean(dA))
        line.append(stats.RMSmean(dA))
        line.append(np.nanstd(dA))

        line.append(np.nanmean(dB))
        line.append(stats.RMSmean(dB))
        line.append(np.nanstd(dB))

        line.append(np.nanmean(dC))
        line.append(stats.RMSmean(dC))
        line.append(np.nanstd(dC))

        line.append(np.nanmean(dD))
        line.append(stats.RMSmean(dD))
        line.append(np.nanstd(dD))

        if D2Dn3D:
            line.append(np.nanmean(dDb))
            line.append(stats.RMSmean(dDb))
            line.append(np.nanstd(dDb))

        line.append(stats.RMSmean([stats.RMSmean(dA),stats.RMSmean(dB),stats.RMSmean(dC)]))
        line.append(stats.RMSmean([stats.RMSmean(dA),stats.RMSmean(dB)]))

        LINES_STK.append(line)

    if split != 0:
        LINES_arr = np.vstack(LINES_STK)
        LINES1 = LINES_arr[:,:split]
        LINES2 = np.column_stack((LINES_arr[:,0],LINES_arr[:,split:]))
        return LINES1,LINES2
    else:
        return LINES_STK

def compar_plot(dico_list_in, namest = 0, namend    = 10  ,
                alpha = 0.8 , diapt = 1.5 , new_style = True,
                colormap = 'gnuplot'):


    if not new_style:
        fig , ax_arr = plt.subplots(2,2)
        fig.set_size_inches(16.53,11.69)
    else:
        fig , ax_arr = plt.subplots(4,1)
        #fig.set_size_inches(11.69,16.53) ### A3
        fig.set_size_inches(8.27,11.69) ### A4

        

    cm = plt.get_cmap(colormap)
    NUM_COLORS = len(dico_list_in)
    color_list = [cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)]


    for color , D in zip(color_list,dico_list_in):

        nametmp  = D['name']
        nametmp  = nametmp.ljust(namend - namest)
        coortype = D['coortype']
        Dtype    = D['Dtype']

        if coortype == 'ENU':
            Ati = '\u0394' + 'East component'
            Bti = '\u0394' + 'North component'
            Cti = '\u0394' + 'Up component'
            Dti = '\u0394' + 'Distance (' + Dtype + ') component'
            ylbl = 'meters'
        elif coortype == 'XYZ':
            Ati = 'delta X'
            Bti = 'delta Y'
            Cti = 'delta Z'
            Dti = 'delta D ' + Dtype
            ylbl = 'meters'
        elif coortype == 'FLH':
            Ati = 'delta Latitude'
            Bti = 'delta Longitude'
            Cti = 'delta Height'
            Dti = 'delta D ' + Dtype + ' (no sens)'
            ylbl = 'degrees'
        elif coortype == 'UTM':
            Ati = 'delta East UTM'
            Bti = 'delta North UTM'
            Cti = 'delta Height'
            Dti = 'delta D UTM ' + Dtype
            ylbl = 'meters'

        else:
            Ati = 'delta A'
            Bti = 'delta B'
            Cti = 'delta C'
            Dti = 'delta D ' + Dtype
            ylbl = '???'

        if not new_style:
            axA = ax_arr[0][0]
            axB = ax_arr[0][1]
            axC = ax_arr[1][0]
            axD = ax_arr[1][1]
        else:
            axA = ax_arr[0]
            axB = ax_arr[1]
            axC = ax_arr[2]
            axD = ax_arr[3]

        axA.plot(D['TA'],D['dA'], '.-' ,label=nametmp[namest:namend] ,
                 markersize=diapt, alpha=alpha,color=color)
        axA.set_ylabel(ylbl)
        axA.legend()
        axA.set_title(Ati)

        axB.plot(D['TB'] ,D['dB'] ,'.-',label=nametmp[namest:namend] ,
                 markersize=diapt, alpha=alpha,color=color)
        axB.set_ylabel(ylbl)
        axB.legend()
        axB.set_title(Bti)

        axC.plot(D['TC'] ,D['dC'],'.-',label=nametmp[namest:namend] ,
                 markersize=diapt, alpha=alpha,color=color)
        axC.set_ylabel(ylbl)
        axC.legend()
        axC.set_title(Cti)

        axD.plot(D['TD'] ,D['dD'],'.-',label=nametmp[namest:namend] ,
                 markersize=diapt, alpha=alpha,color=color)
        axD.set_ylabel(ylbl)
        axD.legend()
        axD.set_title(Dti)
        axD.set_xlabel('Date')

        fig.autofmt_xdate()        
        fig.tight_layout()
        plt.subplots_adjust(top=0.93)
        
    return fig





def compar(tstup , coortype='ENU' , seuil=3. , win=[] , mode='keep' ,
           Dtype='2D3D', namest=0,namend=10,alpha = 0.8 , diapt = 5 , 
           verbose=True, print_report=True,plot=True,interp=True):
    """
        si seuil == 0, pas de nettoyage

        On preconise compar en mode keep :
        Ainsi le Tref est strictement cantonné aux bonnes valeurs des Ti
        En mode del, les Ti sont extrapolé aux valeur de Tref => nan => pbs pour la mad


        if 2D3D /3D2D : selectionne 2D ou 3D (selon) en prioritaire pour dD mais
        détermine 2D et 3D et les introduit dans des clés adhoc dD2D dD3D
        '''
    """

    if len(Dtype) == 4:
        D2n3 = True
        Dtype = Dtype[0:2]
    else:
        D2n3 = False


    if len(tstup) <= 1:
        log.error("len(tstup) <= 1 , do not compare a single element !! ;) ")

    if verbose:
        log.info("COMPARISON SUMMARY")
        log.info("reference : %s",  tstup[0].name)
        log.info("coordonnées : %s" , coortype)

    dicolist = []

    for ivar,tsvar in enumerate(tstup):
        #dico de sortie
        dicovar = dict()

        if verbose:
            log.info("===========================================================")
            log.info(tsvar.name)
            log.info("------------------------------")


#        tsout = TimeSeriePoint()
#        tsout = copy.copy(tsvar)
#        tsout.del_data()


        if tsvar.bool_interp_uptodate == False:
            tsvar.interp_set()

        A, B, C, T, sA , sB , sC = tsvar.to_list(coortype=coortype)

        tsref = tstup[0].interp_get(T,coortype=coortype)
        Aref, Bref, Cref , Tref , sA , sB , sC = tsref.to_list(coortype=coortype)
        
        


        # WTF IS THOSE LINES
        # autowin = time_gap(tstup[0],marge=1.1,mode='keep')
        # win = autowin
        #
        # outbool = []


        # (winlis == [] or (ivar in winlis)) and
        if win != []:
            if verbose:
                log.info("Windowing")
            bl = time_win_T(T,win,mode=mode)
            if verbose:
                log.info("# pts total: %s", len(bl))
                log.info("# pts valid: %s", np.sum(bl))

        else:
            bl = np.asarray([True] * len(Tref))

        dA = A[bl] - Aref[bl]
        dB = B[bl] - Bref[bl]
        dC = C[bl] - Cref[bl]
        Tref = Tref[bl]

        #print "N of diff close to 0 => Bug ..."
        #print 'dA' , np.sum(np.isclose(dA,0))
        #print 'dB' , np.sum(np.isclose(dB,0))
        #print 'dC' , np.sum(np.isclose(dC,0))


        dD2D = np.sqrt(dA ** 2 + dB ** 2)
        dD3D = np.sqrt(dA ** 2 + dB ** 2 + dC ** 2)

        if Dtype == '2D':
            dD = np.sqrt(dA ** 2 + dB ** 2)
        elif Dtype == '3D':
            dD = np.sqrt(dA ** 2 + dB ** 2 + dC ** 2)

        else:
            log.error("wrong Dtype")


#        print "Stats sans Nettoyage"
#        print "------------------------------"
#        print4compar(dA,dB,dC,dD,coortype)

        # nettoyage par la MAD
        if seuil != 0:
            if verbose:
                log.info("MAD cleaning")
                
            dAin = dA
            dA,bb = stats.outlier_mad(dA,threshold=seuil)
            #dAout = dA
            TA = conv.posix2dt(np.array(Tref[bb]))

            dBin = dB
            dB,bb = stats.outlier_mad(dB,threshold=seuil)
            #dBout = dB
            TB = conv.posix2dt(np.array(Tref[bb]))

            dCin = dC
            dC,bb = stats.outlier_mad(dC,threshold=seuil)
            #dCout = dC
            TC = conv.posix2dt(np.array(Tref[bb]))

            dDin = dD
            dD,bb = stats.outlier_mad(dD,threshold=seuil)
            TD = conv.posix2dt(np.array(Tref[bb]))
            #dDout = dD
            if D2n3:
                dD2Din = dD2D
                dD2D,bb = stats.outlier_mad(dD2D,threshold=seuil)
                TD2D = conv.posix2dt(np.array(Tref[bb]))
                #dD2Dout = dD2D
                dD3Din = dD3D
                dD3D,bb = stats.outlier_mad(dD3D,threshold=seuil)
                TD3D = conv.posix2dt(np.array(Tref[bb]))
                #dD3Dout = dD3D

            if verbose:
                log.info("Stats after cleaning")

            if print_report:
                print4compar(dA,dB,dC,dD,coortype)

        else:
            Tdt = conv.posix2dt(Tref)
            TA,TB,TC,TD = Tdt,Tdt,Tdt,Tdt
            dAin,dBin,dCin,dDin = dA,dB,dC,dD

        # pour la legende : les indices de debut et fin du nom

        # fabrication du Dico de sortie
        dicovar['name']     = tsvar.name
        dicovar['coortype'] = coortype
        dicovar['Dtype']    = Dtype

        dicovar['TA'] = np.array(TA)
        dicovar['dA'] = np.array(dA)
        dicovar['dAbrut'] = np.array(dAin)
        dicovar['TB'] = np.array(TB)
        dicovar['dB'] = np.array(dB)
        dicovar['dBbrut'] = np.array(dBin)
        dicovar['TC'] = np.array(TC)
        dicovar['dC'] = np.array(dC)
        dicovar['dCbrut'] = np.array(dCin)
        dicovar['TD'] = np.array(TD)
        dicovar['dD'] = np.array(dD)
        dicovar['dDbrut'] = np.array(dDin)
        dicovar['dDtype'] = np.array(Dtype)
        if D2n3:
            dicovar['TD2D'] = np.array(TD2D)
            dicovar['dD2D'] = np.array(dD2D)
            dicovar['dD2Dbrut'] = np.array(dD2Din)
            dicovar['dD2Dtype'] = np.array('2D')

            dicovar['TD3D'] = np.array(TD3D)
            dicovar['dD3D'] = np.array(dD2D)
            dicovar['dD3Dbrut'] = np.array(dD3Din)
            dicovar['dD3Dtype'] = np.array('3D')

        dicolist.append(dicovar)

    if plot:
        compar_plot(dicolist,namest,namend,alpha,diapt)
        try:
            suptit = ' '.join((tstup[-1].stat , str(tstup[-1].startdate()) ,
                               str(tstup[-1].enddate())))
            plt.suptitle(suptit)
        except:
            pass

    return dicolist

def compar2(tstup,coortype='ENU',seuil=3.,win=[],mode='keep',
           Dtype='3D',namest=0,namend=10,alpha = 5 , diapt = 5 ,verbose=True,
           print_report=True,plot=True):
    """
    160903 cette fonction semble discontinuée
    """

    tsref = tstup[0]
    dicolist = []

    for tsvar in tstup:
        Aref,Bref,Cref,Tref,_,_,_  = tsref.to_list()
        Avar,Bvar,Cvar,Tvar,_,_,_  = tsvar.to_list()

        Tref = np.round(Tref,1)
        Tvar = np.round(Tvar,1)

        Tcommon    = np.intersect1d(Tref,Tvar)
        Tcommon_dt = conv.posix2dt(Tcommon)

        ind_ref = np.nonzero(np.in1d(Tref , Tcommon))[0]
        ind_var = np.nonzero(np.in1d(Tvar , Tcommon))[0]

        dA = Avar[ind_var] - Aref[ind_ref]
        dB = Bvar[ind_var] - Bref[ind_ref]
        dC = Cvar[ind_var] - Cref[ind_ref]

        if Dtype == '2D':
            dD = np.sqrt(dA ** 2 + dB ** 2)
        elif Dtype == '3D':
            dD = np.sqrt(dA ** 2 + dB ** 2 + dC ** 2)

        dicovar = dict()

        dicovar['name']     = tsvar.name
        dicovar['coortype'] = coortype
        dicovar['Dtype']    = Dtype

        dicovar['TA'] = Tcommon_dt
        dicovar['dA'] = np.array(dA)
        dicovar['dAbrut'] = np.array(dA)
        dicovar['TB'] = Tcommon_dt
        dicovar['dB'] = np.array(dB)
        dicovar['dBbrut'] = np.array(dB)
        dicovar['TC'] = Tcommon_dt
        dicovar['dC'] = np.array(dC)
        dicovar['dCbrut'] = np.array(dC)
        dicovar['TD'] = Tcommon_dt
        dicovar['dD'] = np.array(dD)
        dicovar['dDbrut'] = np.array(dD)
        dicovar['dDtype'] = np.array(Dtype)

        dicolist.append(dicovar)

    if plot:
        compar_plot(dicolist,namest,namend,alpha,diapt)
        try:
            suptit = ' '.join((tstup[-1].stat , str(tstup[-1].startdate()) ,
                               str(tstup[-1].enddate())))
            plt.suptitle(suptit)
        except:
            pass

    return dicolist


def round_time(tsin,round_to,
               mode='round'):
    tsout = copy.deepcopy(tsin)
    Tdtin = tsin.to_list(coortype=tsin.initype(),
                         time_as_datetime=True)[3]
    
    Tdtout = conv.round_dt(Tdtin,round_to=round_to,mode=mode,
                           python_dt_out=True)

    tsout.bool_interp_uptodate = False
    
    for pt,t in zip(tsout,Tdtout):
        pt.Tset(conv.numpy_dt2dt(t))
        
    tsout.interp_set()
        
    return tsout

def mad_cleaner(tsin,seuil=3.5,method='dist',coortype='ABC',
                detrend_first=False,output_detrended=False, verbose=False):
    '''
    method : methode d'élimination :
            dist : on élimine les point qu sont trop loin en distance de la posi de ref
            indep : on traite les point independaments

            dist est a privilégier
    output_detrended ne marche que si detrend_first est activé
    '''

    if coortype == 'ABC':
        coortype = tsin.initype()
    log.info(coortype)

    if detrend_first:
        tswork = detrend_ts(tsin,coortype)
    else:
        tswork = tsin

    A,B,C,T,sA,sB,sC = tswork.to_list(coortype)

    D = np.sqrt(A ** 2 + B ** 2 + C ** 2)

    if method == 'dist':
        Dout , bb = stats.outlier_mad(D,seuil)
    if method == 'indep':
        Aout , bbA = stats.outlier_mad(A,seuil)
        Bout , bbB = stats.outlier_mad(B,seuil)
        Cout , bbC = stats.outlier_mad(C,seuil)

        bb = bbA * bbB * bbC

    if output_detrended:
        tsout = bool_cleaner(tswork,bb)
    else:
        tsout = bool_cleaner(tsin,bb)

    if verbose:
        killratio = np.round(float(tsout.nbpts) / float(tsin.nbpts),4)
        log.info("%s, clean ratio : %s, %s, %s" , tsin.stat , tsin.nbpts , tsout.nbpts , killratio)
    return tsout

def sigma_cleaner(tsin,seuil=3,coortype='ABC',cleantype='any', verbose=False):

    if coortype == 'ABC':
        coortype = tsin.initype()

    A,B,C,T,sA,sB,sC = tsin.to_list(coortype)

    sAout , bbsA = stats.outlier_sigma(sA,seuil)
    sBout , bbsB = stats.outlier_sigma(sB,seuil)
    sCout , bbsC = stats.outlier_sigma(sC,seuil)

    if cleantype == 'any':
        boolbad = bbsA * bbsB * bbsC
    elif cleantype == 'all':
        boolbad = bbsA + bbsB + bbsC
    tsout = bool_cleaner(tsin,boolbad)

    if verbose:
        killratio = np.round(float(tsout.nbpts) / float(tsin.nbpts),4)
        log.info("%s, clean ratio : %s, %s, %s" , tsin.stat , tsin.nbpts , tsout.nbpts , killratio)

    return tsout

def std_dev_cleaner(tsin,stddev_threshold,coortype="ABC",cleantype="any",
                    verbose=False):
    """
    A rebooted (1807) version of sigma_cleaner
    just remove values in a timeserie with a high sigma/std deviation
    """

    if coortype == 'ABC':
        coortype = tsin.initype()

    A,B,C,T,sA,sB,sC = tsin.to_list(coortype)

    bbsA = sA <= stddev_threshold
    bbsB = sB <= stddev_threshold
    bbsC = sC <= stddev_threshold

    if cleantype == 'any':
        boolbad = bbsA * bbsB * bbsC
    elif cleantype == 'all':
        boolbad = bbsA + bbsB + bbsC
    tsout = bool_cleaner(tsin,boolbad)

    if verbose:
        killratio = np.round(float(tsout.nbpts) / float(tsin.nbpts),4)

        log.info("%s, clean ratio : %s, %s, %s" , tsin.stat , tsin.nbpts , tsout.nbpts , killratio)

    return tsout

def bool_cleaner(tsin,boollist, verbose=False):
    ''' A partir d'une liste de bool de meme longeur que le nbre de points
        on ne conserve que les points True '''
    tsout = copy.copy(tsin)
    ptslistin = tsin.pts
    if len(ptslistin) != len(boollist):
        log.error("len bool != len pts")
        return 0
    ptslistout = []
    for pt,b in zip(ptslistin,boollist):
         if b:
             ptslistout.append(pt)
    tsout.pts = ptslistout

    if verbose:
        killratio = np.round(float(tsout.nbpts) / float(tsin.nbpts),4)

        log.info("%s, clean ratio : %s, %s, %s" , tsin.stat , tsin.nbpts , tsout.nbpts , killratio)

    return tsout

def linear_regress_find_coeff(tsin,coortype='ENU'):
    A, B, C , T , sA , sB , sC = tsin.to_list(coortype)
    outW = []
    for i,composante in enumerate((A,B,C)):
        M = np.array([ T , np.ones(len(T)) ])
        w  = scipy.linalg.lstsq(M.T,composante)[0]
        outW.append((w[0],w[1]))
    return outW

def detrend_ts(tsin,coortype='ENU',t_origin=None):
    A, B, C , T , sA , sB , sC = tsin.to_list(coortype)
    try:
        Xref , Yref , Zref = tsin.mean_posi()
    except:
        Xref , Yref , Zref = 0,0,0

    W = linear_regress_find_coeff(tsin,coortype)
    UnTr = [] # UnTr
    for i,composante in enumerate((A,B,C)):
        w = W[i]
        comp_line = w[0]*T+w[1]
        comp_untrend = composante - comp_line
        UnTr.append(comp_untrend)

    tsout = ts_from_list(UnTr[0],UnTr[1],UnTr[2],T,coortype,sA,sB,sC,tsin.stat,tsin.name)
    tsout.anex['Xref'] = Xref
    tsout.anex['Yref'] = Yref
    tsout.anex['Zref'] = Zref

    return tsout

def retrend_ts(tsin,a_coef,b_coef,coortype='ENU',t_origin=None):
    """
    a_coefs,b_coefs = 3-tuple/list for the 3 component
    a_coef = m/s
    """
    
    A, B, C , T , sA , sB , sC = tsin.to_list(coortype)
    
    if not t_origin:
        t_origin_use = T[0]
    else:
        t_origin_use = t_origin
        
        
    if not type(t_origin_use) in (float,np.float64):
        print(type(t_origin_use))
        t_origin_use = conv.dt2posix(t_origin_use)

    W = linear_regress_find_coeff(tsin,coortype)
    ReTr = []
    
    for i,composante in enumerate((A,B,C)):
        comp_line = a_coef[i]*(T - t_origin_use) + b_coef[i]
        comp_retrend = composante + comp_line
        ReTr.append(comp_retrend)

    tsout = ts_from_list(ReTr[0],ReTr[1],ReTr[2],
                         T,coortype,
                         sA,sB,sC,
                         tsin.stat,tsin.name)

    return tsout




def linear_regress_ts(tsin,coortype='ENU',titledetails = ''):
    """ doit être cablé ASAP linear_regress_find_coeff"""
    try:
        A, B, C , T , sA , sB , sC = tsin.to_list(coortype)
        fig , axes = plt.subplots(2,2)
        fig.suptitle(tsin.stat + ' ' + titledetails)
        axes = np.reshape(axes,-1)

        W = linear_regress_find_coeff(tsin,coortype)

        for i,composante in enumerate((A,B,C)):
            M = np.array([ T , np.ones(len(T)) ])
#            w  = scipy.linalg.lstsq(M.T,composante)[0]
            w = W[i]
            xi = np.arange(min(T),max(T),10**5)
            line = w[0]*xi+w[1]
            axes[i].plot(conv.posix2dt(xi),line,'r-')
            axes[i].plot(conv.posix2dt(T),composante,'+')
            axes[i].set_title(coortype[i])
            axes[i].set_xlabel('Date')
            axes[i].set_ylabel('Displacement (m)')
            v = w[0] * 31556926 / 10**-3
            v2 = round(v,2)
            log.info('composante %s = %s mm/yr', coortype[i], str(v) )
            
            axes[i].text(0.9, 0.1, str(v2) + 'mm/year' , 
                         horizontalalignment='center',
                         verticalalignment='center',
                         transform=axes[i].transAxes)
    except:
        pass

def linear_regress_ts_discont(tsin,coortype = 'ENU') :
    discont_improved = tsin.discont + [dt.datetime(2099,1,1)]
    for j in range(len(discont_improved)-1):
        log.info('period %s: %s %s',j,discont_improved[j],discont_improved[j+1])
        tsbis = time_win(tsin,[[discont_improved[j],discont_improved[j+1]]])
        datetitle = ' : ' + str(discont_improved[j]) + "\u2192" + str(discont_improved[j+1])
        linear_regress_ts(tsbis,coortype,titledetails=datetitle)


def decimate_cleaner(tsin,minval,in_place = False):
    """
    in_place DOES'NT WORK !!!
    """
    if not in_place:
        tsout = copy.deepcopy(tsin)
    else:
        tsout = tsin
    if minval == 0:
        return tsout

    tsout.del_data()
    for pt in tsin.pts:
        #print pt.T ,minval , np.mod(pt.T,minval)
        if np.mod(np.round(pt.T,2),minval) < 10**-3:
            tsout.add_point(pt)
    tsout.interval_nominal()
    return tsout

def decimate_cleaner_2(tsin,N,in_place = False):
    """ keep a value every N vals """
    if in_place:
        tsout = copy.deepcopy(tsin)
    else:
        tsout = tsin
    tsout.del_data()
    tsout.pts = tsin.pts[:-1:N]
    tsout.interval_nominal()
    return tsout


def mean_list_of_pts(ptslisin):
    """ useful for merge fct
    ONLY IMPLEMENTED FOR ENU coords for the moment """
    E = np.nanmean([p.E for p in ptslisin])
    N = np.nanmean([p.N for p in ptslisin])
    U = np.nanmean([p.U for p in ptslisin])
    sE = np.nanmean([p.sE for p in ptslisin])
    sN = np.nanmean([p.sN for p in ptslisin])
    sU = np.nanmean([p.sU for p in ptslisin])

    Tlis = [p.T for p in ptslisin]
    T = ((np.max(Tlis) - np.min(Tlis)) / 2.) + np.min(Tlis)
    pt = time_series.Point(E,N,U,T,'ENU',sE,sN,sU,ptslisin[0].name)

    return pt


def merge(tsin,N):
    """ merge N points in one """
    tsin.sort()
    tsout = copy.deepcopy(tsin)
    tsout.del_data()
    slic = utils.sliceIt(tsin.pts,N)
    for sl in slic:
        pt = mean_list_of_pts(sl)
        tsout.add_point(pt)
    tsout.interval_nominal()
    return tsout

def merge_ts(ts_list_in):
    pts_list_merged = []
    for ts in ts_list_in:
        pts_list_merged = pts_list_merged + ts.pts

    ts_out = time_series.TimeSeriePoint()
    ts_out.pts = pts_list_merged
    ts_out.anex = ts_list_in[0].anex
    ts_out.interval_nominal()
    ts_out.sort()
    return ts_out


def time_win(tsin,windows,mode='keep',outbool=False):
    if type(windows[0][0]) == dt.datetime:
        Ttype = 'Tdt'
    else:
        Ttype = 'T'

    tsout = copy.copy(tsin)
    tsout.del_data()

    if isinstance(tsout,time_series.TimeSeriePoint):
        data = 'pts'
    elif isinstance(tsout,time_series.TimeSerieObs):
        data = 'obs'
    else:
        log.error('unknown error')

    outboollist = []
    
    for i,e in enumerate(windows):
        log.info('windows %s', i)
        log.info("%s %s",e[0] , e[1])


    for pt in getattr(tsin,data):
        boollist = []
        for win in windows:
            if win[1] < win[0]:
                log.warning("period start > period end")
                log.info("%s %s", win[0] , win[1])
                log.info("%s",mode)

            if win[0] < getattr(pt,Ttype) and win[1] > getattr(pt,Ttype):
                boollist.append(True)
            else:
                boollist.append(False)

        if mode == 'keep' :
            finalbool = np.any(boollist)
        elif mode == 'del':
            finalbool = not np.any(boollist)

        if finalbool:
            tsout.add_point(pt)

        outboollist.append(finalbool)

    tsout.bool_interp_uptodate = False

    #tsout.interp_set()
    if outbool:
        outboollist = np.array(outboollist)
        return tsout,outboollist
    else:
        return tsout

def time_win_T(Tin,win,mode='del'):

    boolfinalist = []

    for t in Tin:
        booltemp = []
        for w in win:
            startw = w[0]
            endw = w[1]

            # CONVENTION : VRAI SI DANS L'INTERVALLE
            # QU'IMPORTE SI KEEP OU DEL
            if startw < t and t < endw :
                booltemp.append(True)
            else:
                booltemp.append(False)

        if mode == 'keep' :
            finalbool = np.any(booltemp)
        elif mode == 'del':
            finalbool = not np.any(booltemp)

        boolfinalist.append(finalbool)

    return np.array(boolfinalist)

def time_gap(tsin,marge=2,mode='del'):
    ''' ENTREE une TimeSerie
        SORTIE une window (liste de listes)'''

    inomi = tsin.i_nomi

    T = tsin.to_list()[3]

    Tdiff = np.diff(np.sort(T))

    boollist = []
    # Dans boollist :
    # False si c'est gap
    # True sinon

    for td in Tdiff:
        if td > inomi * marge:
            boollist.append(False)
        else:
            boollist.append(True)

    # Difference suivant le mode
    if mode == 'keep' :
        modbool = True
    elif mode == 'del':
        modbool = False

    indbool = np.where(np.array(boollist) == modbool)[0]
    protowin = reffram.group_consecutives(indbool)

    winout = []

    for e in protowin:
        if len(e) == 1:
            eout = [T[e[0]],T[e[0]+1]]
        elif len(e) == 2:
            eout = [T[e[0]],T[e[1]+1]]
        else:
            "BUG : time_gap"
        winout.append(eout)

    return winout

def compar_elts_in_ts(ts1,ts2):
    ''' ts2 must contains less elts than ts1 (ts2 = cleaned one)'''
    T1 = ts1.to_list()[3]
    T2 = ts1.to_list()[3]

    for t2 in T2:
        T1.remove(t2)

    return T1


def rotate_pt_cls_solo(tsattin,pointin,Rtype='R1',
                       xyzreftuple = ([1, 0, 0], [0, 1, 0], [0, 0, 1]),
                       angtype='deg'):
    ''' ENTREE : tsattin : une TS d'attitude  (N angles)
                 pointin : UN Point en entrée

        SORTIE : une TSpoint de N points '''

    attlisttmp = tsattin.to_list()
    R = attlisttmp[0]
    P = attlisttmp[1]
    Y = attlisttmp[2]
    T = attlisttmp[3]

    A,B,C,_,_ = pointin()

    ptrotlis = XXX.rotate_points(R,P,Y,[np.array([A,B,C])],
                                 Rtype,xyzreftuple,angtype)

    ptrotlis = utils.shrink_listoflist(ptrotlis)

    tsout = time_series.TimeSeriePoint()

    for p,t in zip(ptrotlis,T):
        #print p
        tsout.add_point(time_series.Point(p[0],p[1],p[2],t,
                                          initype=pointin.initype))

    return tsout


def rotate_points_class(tsattin,ptslin, Rtype='R1',
                        xyzreftuple = ([1, 0, 0], [0, 1, 0], [0, 0, 1]),
                        angtype='deg'):

    listsout = []

    for pt in ptslin:
        ts = rotate_pt_cls_solo(tsattin,pt,Rtype,xyzreftuple,angtype)
        listsout.append(ts)

    return listsout

def add_offset_point(ptin,dA,dB,dC,coortype='ENU'):
    """
    ONLY IMPLEMENTED FOR ENU FOR THE MOMENT
    150415 : remark still necessary ???

    coortype == 'UXYZ' :
        specific case where we correct an Up offset directly in the XYZ coords
        very usefull for an antenna offset in for a moving GPS
        (but works only for the up)
    """
    ptout = copy.copy(ptin)
    if coortype == 'ENU':
        ptout.ENUset(ptin.E + dA,ptin.N + dB,ptin.U + dC,
                     ptin.sE,ptin.sN,ptin.sU)
    elif coortype == 'XYZ':
        ptout.XYZset(ptin.X + dA,ptin.Y + dB,ptin.Z + dC,
                     ptin.sX,ptin.sY,ptin.sZ)
    elif coortype == 'FLH':
        ptout.FLHset(ptin.F + dA,ptin.L + dB,ptin.H + dC,
                     ptin.sF,ptin.sL,ptin.sH)

    elif coortype == 'UXYZ':
        N = conv.normal_vector(ptin.F , ptin.L , ptin.H)
        dA2 , dB2 , dC2 = N * dC
        ptout.XYZset(ptin.X + dA2,ptin.Y + dB2,ptin.Z + dC2,
                     ptin.sX,ptin.sY,ptin.sZ)

    return ptout

def add_offset_ts(tsin,dA,dB,dC,coortype='ENU'):
    """
    return a copy of the tsin
    (tsin won't be affected)

    coortype == 'UXYZ' :
        specific case where we correct an Up offset directly in the XYZ coords
        very usefull for an antenna offset in for a moving GPS
        (but works only for the up)
    """
    tsout = copy.deepcopy(tsin)
    tsout.del_data()
    newpts = [add_offset_point(pt,dA,dB,dC,coortype) for pt in tsin.pts]
    for pt in newpts:
        tsout.add_point(pt)
    return tsout

def add_offset_smart_for_GINS_kine(tsin,tslist_offset_3ple,list_windows,coortype='XYZ'):
    """
    tslist_offset_3ple : list of len N containing (dX,dY,dZ) offsets
    list_windows       : list of len N-1 containing dates of changes
    """

    if type(list_windows[0]) is int:
        list_windows = [conv.posix2dt(e) for e in list_windows]

    list_windows = [dt.datetime(1980,1,1)] + list_windows + [dt.datetime(2099,1,1)]
    tsout = copy.deepcopy(tsin)
    tsout.del_data()
    olddX ,olddY , olddZ = 12345 , 12345 , 12345
    for pt in tsin:
        for i in range(len(list_windows)-1):
            if list_windows[i] <= pt.Tdt < list_windows[i+1]:
                curdX , curdY , curdZ = tslist_offset_3ple[i][0],tslist_offset_3ple[i][1],tslist_offset_3ple[i][2]
                if curdX != olddX or curdY != olddY or curdZ != olddZ:
                    olddX ,olddY , olddZ = curdX , curdY , curdZ
                    log.info('change of time window / offset')
                ptout = add_offset_point(pt,curdX,curdY,curdZ,coortype='XYZ')
                tsout.add_point(ptout)
                continue
    return tsout


def find_pts_from_ts_with_time(tin,tstupin,tol=0.001):
    ptsout = []
    for ts in tstupin:
        pt, i = ts.find_point(tin,tol=tol)
        if np.isnan(i):
            log.warning("no point found")
        ptsout.append(pt)
    return ptsout

def dist_btwn_2pts(ptA,ptB,coortype='XYZ'):

        A = np.array([ptA.X,ptA.Y,ptA.Z])
        B = np.array([ptB.X,ptB.Y,ptB.Z])

        return conv.dist(A,B)

def dist_diff_btwn_2pts(ptA,ptB):
    #dérive la distance entre 2 points A et B
        A = np.array([ptA.X,ptA.Y,ptA.Z])
        B = np.array([ptB.X,ptB.Y,ptB.Z])

        dAB   = A-B
        dist  = np.linalg.norm(dAB)

        diffA =   dAB / dist
        diffB = - dAB / dist

        return diffA, diffB

def helmert_trans(tsin,params='itrf2008_2_etrf2000', invert=False):
    tsout = copy.deepcopy(tsin)
    [pt.helmert_trans(params,invert) for pt in tsout.pts]
    return tsout


def velocity_trans(tsin , vx, vy, vz, epoc_init = 'auto' , epoc_end = 'auto'):
    tsout = copy.deepcopy(tsin)
    [pt.velocity_trans(vx, vy, vz, epoc_init , epoc_end) for pt in tsout.pts]
    return tsout


def interpolator_light(T,X,Y,Z):
    from scipy.interpolate import interp1d

    IX = interp1d(T,X,bounds_error=False,fill_value=np.nan)
    IY = interp1d(T,Y,bounds_error=False,fill_value=np.nan)
    IZ = interp1d(T,Z,bounds_error=False,fill_value=np.nan)

    return IX , IY , IZ

def interpolator_with_extrapolated(T,X,Y,Z):
    from scipy.interpolate import interp1d

    IX = interp1d(T,X,bounds_error=False,fill_value="extrapolate")
    IY = interp1d(T,Y,bounds_error=False,fill_value="extrapolate")
    IZ = interp1d(T,Z,bounds_error=False,fill_value="extrapolate")

    return IX , IY , IZ

def mean_posi_multi(tstup):
    mp = tstup[0].mean_posi()
    for ts in tstup:
        ts.ENUcalc(mp)
    return None

def refENU_for_tslist(tslist_in, tsref_marker = 0):
    """
    tsref_marker : indice of the reference time serie OR the 'all' keyword
    in this case all the time series mean position will be averaged
    """
    if tsref_marker == 'all':
        refENU_stk =  []
        for ts in tslist_in:
            refENUtmp = ts.mean_posi()
            refENU_stk.append(refENUtmp)

        X = np.mean([renu.X for renu in refENU_stk])
        Y = np.mean([renu.Y for renu in refENU_stk])
        Z = np.mean([renu.Z for renu in refENU_stk])

        refENU = time_series.Point(X,Y,Z,0,initype='XYZ')

    else:
        refENU = tslist_in[tsref_marker].mean_posi()


    for ts in tslist_in:
        ts.ENUcalc(refENU)

    return refENU

def time_win_multi(inplis):
    strt = inplis[0].startdate()
    end  = inplis[0].enddate()
    tsoutlis = [inplis[0]]
    for ts in list(inplis)[1:]:
        tsout = time_win(ts,[[strt,end]])
        tsoutlis.append(tsout)
    return tsoutlis

def ts_from_list(A,B,C,T,initype,sA=[],sB=[],sC=[],stat='STAT',name='NoName'):
    tsout = time_series.TimeSeriePoint()
    if len(sA) == 0:
        for a,b,c,t in zip(A,B,C,T):
            pt = time_series.Point(a,b,c,t,initype)
            tsout.add_point(pt)
    else:
        for a,b,c,t,sa,sb,sc in zip(A,B,C,T,sA,sB,sC):
            pt = time_series.Point(a,b,c,t,initype,sa,sb,sc)
            tsout.add_point(pt)
    tsout.stat = stat
    tsout.name = name
    if initype == 'ENU':
        tsout.boolENU = True
    else:
        tsout.boolENU = False
    if initype == 'UTM':
        tsout.boolUTM = True
    else:
        tsout.boolUTM = False
    return tsout


def baselines_calc(ts_list,
                   plani_only=False,
                   substract_offset = np.median,
                   symetric_calc  = False,
                   symetric_storage = True):
    """
    

    Parameters
    ----------
    ts_list : list
        list of TimeSeries.
    plani_only : bool, optional
        If True, compute the baseline varation on the East and North 
        component only.
        The default is False.
    substract_offset : function or None, optional
        A function to substract the offset. 
        Can be `np.mean` (substract the mean value), 
        `np.median` (substract the median value), 
        or `lambda x: x[0]` (substract the 1st value)
        The default is np.median.
    symetric_calc: bool, optional
        If True, compute the baseline variation 2 times,
        for stat1 > stat2 and stat2 > stat1
        Might be useful to compare planimetic computation
        The default is False.
    symetric_storage : bool, optional
        If True, store the baseline variation 2 times,
        ``dictstore[stat1][stat2]`` and ``dictstore[stat2][stat1]``
        ``symetric_calc`` overrides this option
        The default is True.

    Returns
    -------
    dictstore : dict
        A dictionnary of dictionnaries of Pandas Series, 
        containing the baseline variations
        e.g. ``dictstore[stat1][stat2] = bl_series``

    """

    sites_store = [ts.stat for ts in ts_list]
    
    if symetric_storage:
        dictstore = {key2: {key: None for key in sites_store} for key2 in sites_store}
    else:
        dictstore = {key2: {} for key2 in sites_store}

    if symetric_calc:
        couples = itertools.permutations(ts_list, 2)
    else:
        couples = itertools.combinations(ts_list, 2)

        
    for ts1orig, ts2orig in couples:
        
        ts1 = copy.deepcopy(ts1orig)
        ts2 = copy.deepcopy(ts2orig)
        
        if plani_only:
            ts1mean = ts1.mean_posi(meanormed='median')
            ts1.ENUcalc(ts1mean)
            ts2.ENUcalc(ts1mean)
        
        def _df_prepa(tsin):
            if not plani_only:
                dfout = tsin.to_dataframe("XYZ")       
                dfout = dfout[["Tdt","X","Y","Z"]]
            else:
                dfout = tsin.to_dataframe("ENU")       
                dfout = dfout[["Tdt","E","N"]]
                
            dfout = dfout.set_index("Tdt")
            return dfout
        
        df1 = _df_prepa(ts1)
        df2 = _df_prepa(ts2)
        
        i1 = df1.index
        i2 = df2.index
        iok = i1.intersection(i2)
        
        df1i = df1.loc[iok]
        df2i = df2.loc[iok]
            
        dfbl = (df1i - df2i).apply(np.linalg.norm, axis=1) 
        
        if len(dfbl) > 0 and substract_offset:
            dfbl = dfbl - substract_offset(dfbl)
          
        
        dictstore[ts1.stat][ts2.stat] = dfbl
        if symetric_storage and not symetric_calc:
            dictstore[ts2.stat][ts1.stat] = dfbl
            
    return dictstore



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.reffram contains functions for operations
related to GNSS-products


it can be imported directly with:
from geodezyx import reffram

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import itertools
import matplotlib
import matplotlib.pyplot as plt
import natsort
import numpy as np
import os 
import pandas as pd
import re

### disabled and imported directly in the needed fct
## import geodezyx.reffram.sofa18 as sofa

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import stats
from geodezyx import utils

#### Import the logger
import logging
log = logging.getLogger(__name__)

##########  END IMPORT  ##########

def compar_orbit(Data_inp_1,Data_inp_2,step_data = 900,
                 sats_used_list = ['G'],
                 name1='',name2='',use_name_1_2_for_table_name = False,
                 RTNoutput = True,convert_ECEF_ECI=True,
                 clean_null_values = True,
                 conv_coef=10**3,return_satNull = False):
    """
    Compares 2 GNSS orbits files (SP3), and gives a summary plot and a
    statistics table

    Parameters
    ----------
    Data_inp_1 & Data_inp_2 : str or Pandas DataFrame
        contains the orbits or path (string) to the sp3

    step_data : int
        per default data sampling

    sats_used_list : list of str
        used constellation or satellite : G E R C ... E01 , G02 ...
        Individuals satellites are prioritary on whole constellations
        e.g. ['G',"E04"]


    RTNoutput : bool
        select output, Radial Transverse Normal or XYZ

    convert_ECEF_ECI : bool
        convert sp3 ECEF => ECI (Terrestrial => Celestrial)
        must be True in operational to avoid artifacts.

    name1 & name2 : str (optionals)
        optional custom names for the 2 orbits

    use_name_1_2_for_table_name : bool
        False : use name 1 and 2 for table name, use datafile instead

    clean_null_values : bool or str
        if True or "all" remove sat position in all X,Y,Z values
        are null (0.000000)
        if "any", remove sat position if X or Y or Z is null
        if False, keep everything
        
    conv_coef : int
        conversion coefficient, km to m 10**3, km to mm 10**6

    Returns
    -------
    Diff_sat_all : Pandas DataFrame
    contains differences b/w Data_inp_1 & Data_inp_2
    in Radial Transverse Normal OR XYZ frame

        Attributes of Diff_sat_all :
            Diff_sat_all.name : title of the table

    Note
    ----
    clean_null_values if useful (and necessary) only if
    convert_ECEF_ECI = False
    if convert_ECEF_ECI = True, the cleaning will be done by
    a side effect trick : the convertion ECEF => ECI will generate NaN
    for a zero-valued position
    But, nevertheless, activating  clean_null_values = True is better
    This Note is in fact usefull if you want to see bad positions on a plot
    => Then convert_ECEF_ECI = False and clean_null_values = False

    References
    ----------
    "Coordinate Systems", ASEN 3200 1/24/06 George H. Born

    """

    # selection of both used Constellations AND satellites
    const_used_list = []
    sv_used_list    = []
    for sat in sats_used_list:
        if len(sat) == 1:
            const_used_list.append(sat)
        elif len(sat) == 3:
            sv_used_list.append(sat)
            if not sat[0] in const_used_list:
                const_used_list.append(sat[0])

    # Read the files or DataFrames
    # metadata attributes are not copied
    # Thus, manual copy ...
    # (Dirty way, should be impoved without so many lines ...)
    if type(Data_inp_1) is str:
        D1orig = files_rw.read_sp3(Data_inp_1,epoch_as_pd_index=True)
    else:
        D1orig = Data_inp_1.copy(True)
        try:
            D1orig.name = Data_inp_1.name
        except:
            D1orig.name = "no_name"
        try:
            D1orig.path = Data_inp_1.path
        except:
            D1orig.path = "no_path"
        try:
            D1orig.filename = Data_inp_1.filename
        except:
            D1orig.filename = "no_filename"

    if type(Data_inp_2) is str:
        D2orig = files_rw.read_sp3(Data_inp_2,epoch_as_pd_index=True)
    else:
        D2orig = Data_inp_2.copy(True)
        try:
            D2orig.name = Data_inp_2.name
        except:
            D2orig.name = "no_name"
        try:
            D2orig.path = Data_inp_2.path
        except:
            D2orig.path = "no_path"
        try:
            D2orig.filename = Data_inp_2.filename
        except:
            D2orig.filename = "no_filename"

    #### NB : It has been decided with GM that the index of a SP3 dataframe
    ####      will be integers, not epoch datetime anymore
    ####      BUT here, for legacy reasons, the index has to be datetime

    if isinstance(D1orig.index[0], (int, np.integer)):
        D1orig.set_index("epoch",inplace=True)

    if isinstance(D2orig.index[0], (int, np.integer)):
        D2orig.set_index("epoch",inplace=True)

    Diff_sat_stk = []

    # This block is for removing null values
    if clean_null_values:
        if clean_null_values == "all":
            all_or_any = np.all
        elif clean_null_values == "any":
            all_or_any = np.any
        else:
            all_or_any = np.all

        xyz_lst = ['x','y','z']

        D1_null_bool = all_or_any(np.isclose(D1orig[xyz_lst],0.),axis=1)
        D2_null_bool = all_or_any(np.isclose(D2orig[xyz_lst],0.),axis=1)

        D1 = D1orig[np.logical_not(D1_null_bool)]
        D2 = D2orig[np.logical_not(D2_null_bool)]

        if np.any(D1_null_bool) or np.any(D2_null_bool):
            sat_nul = utils.join_improved(" " ,*list(set(D1orig[D1_null_bool]["prn"])))
            log.warning("Null values contained in SP3 files : ")
            log.warning("f1: %s %s" , np.sum(D1_null_bool) , utils.join_improved(" " , *list(set(D1orig[D1_null_bool]["prn"]))))
            log.warning("f2: %s %s" , np.sum(D2_null_bool) , utils.join_improved(" " , *list(set(D2orig[D2_null_bool]["prn"]))))
        else:
            sat_nul = []

    else:
        D1 = D1orig.copy()
        D2 = D2orig.copy()

    for constuse in const_used_list:
        D1const = D1[D1['sys'] == constuse]
        D2const = D2[D2['sys'] == constuse]

        # checking if the data correspond to the step
        bool_step1 = np.mod((D1const.index - np.min(D1.index)).seconds,step_data) == 0
        bool_step2 = np.mod((D2const.index - np.min(D2.index)).seconds,step_data) == 0

        D1window = D1const[bool_step1]
        D2window = D2const[bool_step2]
        
        # find common sats and common epochs
        sv_set   = sorted(list(set(D1window['prni']).intersection(set(D2window['prni']))))
        epoc_set = sorted(list(set(D1window.index).intersection(set(D2window.index))))

        # if special selection of sats, then apply it
        # (it is late and this selection is incredibely complicated ...)
        if np.any([True  if constuse in e else False for e in sv_used_list]):
            # first find the selected sats for the good constellation
            sv_used_select_list = [int(e[1:]) for e in sv_used_list if constuse in e]
            #and apply it
            sv_set = sorted(list(set(sv_set).intersection(set(sv_used_select_list))))

        for svv in sv_set:
            # First research : find corresponding epoch for the SV
            # this one is sufficent if there is no gaps (e.g. with 0.00000) i.e.
            # same nb of obs in the 2 files
            # NB : .reindex() is smart, it fills the DataFrame
            # with NaN
            try:
                D1sv_orig = D1window[D1window['prni'] == svv].reindex(epoc_set)
                D2sv_orig = D2window[D2window['prni'] == svv].reindex(epoc_set)
            except Exception as exce:
                log.info("ERR : Unable to re-index with an unique epoch")
                log.info("      are you sure there is no multiple-defined epochs for the same sat ?")
                log.info("      it happens e.g. when multiple ACs are in the same DataFrame ")
                log.info("TIP : Filter the input Dataframe before calling this fct with")
                log.info("      DF = DF[DF['AC'] == 'gbm']")
                
                Dtmp1 = D1orig[D1orig['prni'] == svv]
                Dtmp2 = D2orig[D2orig['prni'] == svv]
                
                dupli1 = np.sum(Dtmp1.duplicated(["epoch","prn"]))
                dupli2 = np.sum(Dtmp2.duplicated(["epoch","prn"]))
                
                log.info("FWIW: duplicated epoch/sat in DF1 & DF2: %s %s",dupli1,dupli2)

                raise exce

            # Second research, it is a security in case of gap
            # This step is useless, because .reindex() will fill the DataFrame
            # with NaN
            if len(D1sv_orig) != len(D2sv_orig):
                log.info("different epochs nbr for SV %s %s %s",svv,len(D1sv_orig),len(D2sv_orig))
                epoc_sv_set = sorted(list(set(D1sv_orig.index).intersection(set(D2sv_orig.index))))
                D1sv = D1sv_orig.loc[epoc_sv_set]
                D2sv = D2sv_orig.loc[epoc_sv_set]
            else:
                D1sv = D1sv_orig
                D2sv = D2sv_orig

            P1     = D1sv[['x','y','z']]
            P2     = D2sv[['x','y','z']]

            # Start ECEF => ECI
            if convert_ECEF_ECI:
                # Backup because the columns xyz will be reaffected
                #D1sv_bkp = D1sv.copy()
                #D2sv_bkp = D2sv.copy()
    
                P1b = conv.ECEF2ECI(np.array(P1),conv.dt_gpstime2dt_utc(P1.index.to_pydatetime(),out_array=True))
                P2b = conv.ECEF2ECI(np.array(P2),conv.dt_gpstime2dt_utc(P2.index.to_pydatetime(),out_array=True))

                D1sv[['x','y','z']] = P1b
                D2sv[['x','y','z']] = P2b

                P1  = D1sv[['x','y','z']]
                P2  = D2sv[['x','y','z']]
            # End ECEF => ECI

            if not RTNoutput:
                # Compatible with the documentation +
                # empirically tested with OV software
                # it is  P1 - P2 (and not P2 - P1)
                Delta_P = P1 - P2


                Diff_sat = Delta_P.copy()
                Diff_sat.columns = ['dx','dy','dz']

            else:
                rnorm = np.linalg.norm(P1,axis=1)

                Vx = utils.diff_pandas(D1sv,'x')
                Vy = utils.diff_pandas(D1sv,'y')
                Vz = utils.diff_pandas(D1sv,'z')

                V =  pd.concat((Vx , Vy , Vz),axis=1)
                V.columns = ['vx','vy','vz']

                R = P1.divide(rnorm,axis=0)
                R.columns = ['xnorm','ynorm','znorm']

                H      = pd.DataFrame(np.cross(R,V),columns=['hx','hy','hz'])
                hnorm  = np.linalg.norm(H,axis=1)

                C         = H.divide(hnorm,axis=0)
                C.columns = ['hxnorm','hynorm','hznorm']

                I         = pd.DataFrame(np.cross(C,R),columns=['ix','iy','iz'])

                R_ar = np.array(R)
                I_ar = np.array(I)
                C_ar = np.array(C)

                #R_ar[1]
                Beta = np.stack((R_ar,I_ar,C_ar),axis=1)

                # Compatible with the documentation +
                # empirically tested with OV software
                # it is  P1 - P2 (and not P2 - P1)
                Delta_P = P1 - P2

                # Final determination
                Astk = []

                for i in range(len(Delta_P)):
                    A = np.dot(Beta[i,:,:],np.array(Delta_P)[i])
                    Astk.append(A)

                Diff_sat = pd.DataFrame(np.vstack(Astk),
                                   index = P1.index,columns=['dr','dt','dn'])

            Diff_sat = Diff_sat * conv_coef # metrer conversion

            Diff_sat['sys'] = [constuse] * len(Diff_sat.index)
            Diff_sat['prni']    = [svv]      * len(Diff_sat.index)
            Diff_sat['prn']   = [constuse + str(svv).zfill(2)] * len(Diff_sat.index)

            Diff_sat_stk.append(Diff_sat)

    Diff_sat_all = pd.concat(Diff_sat_stk)
    Date = Diff_sat.index[0]

    # Attribute definition
    if RTNoutput:
        Diff_sat_all.frame_type = 'RTN'

        # Pandas donesn't manage well iterable as attribute
        # So, it is separated
        Diff_sat_all.frame_col_name1 = 'dr'
        Diff_sat_all.frame_col_name2 = 'dt'
        Diff_sat_all.frame_col_name3 = 'dn'

    else:
        # Pandas donesn't manage well iterable as attribute
        # So, it is separated
        Diff_sat_all.frame_col_name1 = 'dx'
        Diff_sat_all.frame_col_name2 = 'dy'
        Diff_sat_all.frame_col_name3 = 'dz'

        if convert_ECEF_ECI:
            Diff_sat_all.frame_type = 'ECI'
        else:
            Diff_sat_all.frame_type = 'ECEF'


    # Name definitions
    if name1:
        Diff_sat_all.name1 = name1
    else:
        Diff_sat_all.name1 = D1orig.name

    if name2:
        Diff_sat_all.name2 = name2
    else:
        Diff_sat_all.name2 = D2orig.name

    Diff_sat_all.filename1 = D1orig.filename
    Diff_sat_all.filename2 = D2orig.filename

    Diff_sat_all.path1 = D1orig.path
    Diff_sat_all.path2 = D2orig.path

    Diff_sat_all.name = ' '.join(('Orbits comparison ('+Diff_sat_all.frame_type +') b/w',
                                  Diff_sat_all.name1 ,'(ref.) and',
                                  Diff_sat_all.name2 ,',',Date.strftime("%Y-%m-%d"),
                                  ', doy', str(conv.dt2doy(Date))))

    
    if return_satNull:
        return Diff_sat_all, sat_nul
    else:
        return Diff_sat_all


def compar_orbit_plot(Diff_sat_all_df_in,
                      save_plot=False,
                      save_plot_dir="",
                      save_plot_name="auto",
                      save_plot_name_suffix=None,
                      save_plot_ext=(".pdf",".png",".svg"),
                      yaxis_limit=None,
                      yaxis_label_unit="m"):
    """
    General description

    Parameters
    ----------
    Diff_sat_all_df_in : DataFrame
        a DataFrame produced by compar_orbit
        
    yaxis_limit : 3-tuple iterable or 2-element tuple
        force the y axis limits. must look like 
        [(ymin_r,ymax_r),(ymin_t,ymax_t),(ymin_n,ymax_n)]
        to control all the axis independely
        OR
        (ymin,ymax)
        to set all th axis at the same limits 

    Returns
    -------
    the Figure and the 3 Axes if no save is asked
    export path (str) if save is asked
    but plot a plot anyway
    """

    import matplotlib.dates as mdates
    fig,[axr,axt,axn] = plt.subplots(3,1,sharex='all')

    satdispo = natsort.natsorted(list(set(Diff_sat_all_df_in['prn'])))

    SymbStk = []

    cm = plt.get_cmap('viridis')
    NUM_COLORS = len(satdispo)
    Colors = [cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)]

    # Pandas donesn't manage well iterable as attribute
    # So, it is separated
    try:
        col_name0 = Diff_sat_all_df_in.frame_col_name1
        col_name1 = Diff_sat_all_df_in.frame_col_name2
        col_name2 = Diff_sat_all_df_in.frame_col_name3
    except:
        col_name0 = Diff_sat_all_df_in.columns[0]
        col_name1 = Diff_sat_all_df_in.columns[1]
        col_name2 = Diff_sat_all_df_in.columns[2]

    for satuse,color in zip(satdispo,Colors):
        Diffuse = Diff_sat_all_df_in[Diff_sat_all_df_in['prn'] == satuse]

        Time = Diffuse.index
        R    = Diffuse[col_name0]
        T    = Diffuse[col_name1]
        N    = Diffuse[col_name2]

        #fig.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')

        Symb = axr.plot(Time,R,label=satuse,c=color)
        axt.plot(Time,T,label=satuse,c=color)
        axn.plot(Time,N,label=satuse,c=color)

        SymbStk.append(Symb[0])

        fig.autofmt_xdate()


    ylabuni = " (" + yaxis_label_unit + ")"
    
    if Diff_sat_all_df_in.frame_type == 'RTN':
        axr.set_ylabel('Radial diff.'     + ylabuni)
        axt.set_ylabel('Transverse diff.' + ylabuni)
        axn.set_ylabel('Normal diff.'     + ylabuni)

    else:
        axr.set_ylabel(Diff_sat_all_df_in.frame_type + ' X diff.' + ylabuni)
        axt.set_ylabel(Diff_sat_all_df_in.frame_type + ' Y diff.' + ylabuni)
        axn.set_ylabel(Diff_sat_all_df_in.frame_type + ' Z diff.' + ylabuni)


    y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)
    axr.yaxis.set_major_formatter(y_formatter)
    axt.yaxis.set_major_formatter(y_formatter)
    axn.yaxis.set_major_formatter(y_formatter)
    
    if yaxis_limit and len(yaxis_limit) == 3: ### indep. axis limit
        axr.set_ylim(yaxis_limit[0])
        axt.set_ylim(yaxis_limit[1])
        axn.set_ylim(yaxis_limit[2])
    elif yaxis_limit and len(yaxis_limit) == 2:
        axr.set_ylim(yaxis_limit)
        axt.set_ylim(yaxis_limit)
        axn.set_ylim(yaxis_limit)
    else:
        pass
        
        
    import matplotlib.dates as mdates
    fig.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')

    lgd = fig.legend(tuple(SymbStk), satdispo , loc='lower center',ncol=8,
                     columnspacing=1)

    fig.set_size_inches(8.27,11.69)
    plt.suptitle(Diff_sat_all_df_in.name)
    plt.tight_layout()
    plt.subplots_adjust(top=0.95)
    plt.subplots_adjust(bottom=0.15)

    if save_plot:
        if save_plot_name == "auto":
            save_plot_name = "_".join((Diff_sat_all_df_in.name1,
                                      Diff_sat_all_df_in.name2,
                                      Diff_sat_all_df_in.index.min().strftime("%Y-%m-%d")))
            
        if save_plot_name_suffix:
            save_plot_name = save_plot_name + '_' + save_plot_name_suffix

        for ext in save_plot_ext:
            save_plot_path = os.path.join(save_plot_dir,save_plot_name)
            plt.savefig(save_plot_path + ext)
            return_val = save_plot_path
            
    else:
        return_val = fig,(axr,axt,axn)

    return return_val

def compar_orbit_table(Diff_sat_all_df_in,RMS_style = 'natural',
                       light_tab  = False):
    """
    Generate a table with statistical indicators for an orbit comparison
    (RMS mean, standard dev, ...)
    
    Parameters
    ----------
    Diff_sat_all_df_in : Pandas DataFrame
        a DataFrame produced by compar_orbit

    RMS_style : str
        'natural': use the natural definition of the RMS
        'GRGS': RMS calc based on the GRGS definition of the RMS (OV help)
        is actually the standard deviation
        'kouba': RMS as defined in Kouba et al. 1994, p75
        using the degree of freedom (3*Nobs - 7)

    light_tab : bool
        produce a table with only RMS, with min/max/arithmetic instead

    Returns
    -------
    Compar_tab_out : DataFrame
        Statistical results of the comparison

    Note
    ----
    you can pretty print the output DataFrame using tabular module
    here is a template:

    >>> from tabulate import tabulate
    >>> print(tabulate(ComparTable,headers="keys",floatfmt=".4f"))
    """

    sat_list = utils.uniq_and_sort(Diff_sat_all_df_in['prn'])

    # Pandas donesn't manage well iterable as attribute
    # So, it is separated
    try:
        col_name0 = Diff_sat_all_df_in.frame_col_name1
        col_name1 = Diff_sat_all_df_in.frame_col_name2
        col_name2 = Diff_sat_all_df_in.frame_col_name3
    except:
        col_name0 = Diff_sat_all_df_in.columns[0]
        col_name1 = Diff_sat_all_df_in.columns[1]
        col_name2 = Diff_sat_all_df_in.columns[2]

    rms_stk = []

    for sat in sat_list:
        Diffwork = utils.df_sel_val_in_col(Diff_sat_all_df_in,'prn',sat)

        if RMS_style == "natural":
            rms_A = stats.rms_mean(Diffwork[col_name0])
            rms_B = stats.rms_mean(Diffwork[col_name1])
            rms_C = stats.rms_mean(Diffwork[col_name2])
        elif RMS_style == "GRGS":
            rms_A = stats.rms_mean(Diffwork[col_name0] - Diffwork[col_name0].mean())
            rms_B = stats.rms_mean(Diffwork[col_name1] - Diffwork[col_name1].mean())
            rms_C = stats.rms_mean(Diffwork[col_name2] - Diffwork[col_name2].mean())
        elif RMS_style == "kouba":
            rms_A = stats.rms_mean_kouba(Diffwork[col_name0])
            rms_B = stats.rms_mean_kouba(Diffwork[col_name1])
            rms_C = stats.rms_mean_kouba(Diffwork[col_name2])
            
        RMS3D = np.sqrt(rms_A**2 + rms_B**2 + rms_C**2)

        min_A = Diffwork[col_name0].min()
        min_B = Diffwork[col_name1].min()
        min_C = Diffwork[col_name2].min()

        max_A = Diffwork[col_name0].max()
        max_B = Diffwork[col_name1].max()
        max_C = Diffwork[col_name2].max()

        mean_A = Diffwork[col_name0].mean()
        mean_B = Diffwork[col_name1].mean()
        mean_C = Diffwork[col_name2].mean()

        if light_tab:
            rms_stk.append([rms_A,rms_B,rms_C,RMS3D])
        else:
            rms_stk.append([rms_A,rms_B,rms_C,RMS3D,
                            min_A,max_A,mean_A,
                            min_B,max_B,mean_B,
                            min_C,max_C,mean_C])


    #################################
             # ALL SATS
    if RMS_style == "natural":
        rms_A = stats.rms_mean(Diff_sat_all_df_in[col_name0])
        rms_B = stats.rms_mean(Diff_sat_all_df_in[col_name1])
        rms_C = stats.rms_mean(Diff_sat_all_df_in[col_name2])
        RMS3D = np.sqrt(rms_A**2 + rms_B**2 + rms_C**2)
    elif RMS_style == "GRGS":
        rms_A = stats.rms_mean(Diff_sat_all_df_in[col_name0] - Diff_sat_all_df_in[col_name0].mean())
        rms_B = stats.rms_mean(Diff_sat_all_df_in[col_name1] - Diff_sat_all_df_in[col_name1].mean())
        rms_C = stats.rms_mean(Diff_sat_all_df_in[col_name2] - Diff_sat_all_df_in[col_name2].mean())
        RMS3D = np.sqrt(rms_A**2 + rms_B**2 + rms_C**2)
    elif RMS_style == "kouba":
        rms_A = stats.rms_mean_kouba(Diff_sat_all_df_in[col_name0])
        rms_B = stats.rms_mean_kouba(Diff_sat_all_df_in[col_name1])
        rms_C = stats.rms_mean_kouba(Diff_sat_all_df_in[col_name2])
        RMS3D = np.sqrt(rms_A**2 + rms_B**2 + rms_C**2)


    min_A = Diff_sat_all_df_in[col_name0].min()
    min_B = Diff_sat_all_df_in[col_name1].min()
    min_C = Diff_sat_all_df_in[col_name2].min()

    max_A = Diff_sat_all_df_in[col_name0].max()
    max_B = Diff_sat_all_df_in[col_name1].max()
    max_C = Diff_sat_all_df_in[col_name2].max()

    mean_A = Diff_sat_all_df_in[col_name0].mean()
    mean_B = Diff_sat_all_df_in[col_name1].mean()
    mean_C = Diff_sat_all_df_in[col_name2].mean()

    if light_tab:
        rms_stk.append([rms_A,rms_B,rms_C,RMS3D])
    else:
        rms_stk.append([rms_A,rms_B,rms_C,RMS3D,
                        min_A,max_A,mean_A,
                        min_B,max_B,mean_B,
                        min_C,max_C,mean_C])

             # ALL SATS
    #################################

    if  Diff_sat_all_df_in.frame_type == 'RTN':
        if light_tab:
            cols_nam = ["rmsR","rmsT","rmsN","rms3D"]
        else:
            cols_nam = ["rmsR","rmsT","rmsN","rms3D",
                        "minR","maxR","meanR",
                        "minT","maxT","meanT",
                        "minN","maxN","meanN"]

    else:
        if light_tab:
            cols_nam = ["rmsX","rmsY","rmsZ","rms3D"]
        else:
            cols_nam = ["rmsX","rmsY","rmsZ","rms3D",
                        "minX","maxX","meanX",
                        "minY","maxY","meanY",
                        "minZ","maxZ","meanZ"]

    Compar_tab_out     = pd.DataFrame(rms_stk,index=sat_list + ['ALL'],
                                      columns=cols_nam)

    return Compar_tab_out


def compar_orbit_frontend(DataDF1,DataDF2,ac1,ac2, sats_used_list = ['G']):
    K = compar_orbit(DataDF1[DataDF1["ac"] == ac1],
                     DataDF2[DataDF2["ac"] == ac2],
                     sats_used_list=sats_used_list)
    compar_orbit_plot(K)
    return K





def compar_clock(DFclk_inp_1,DFclk_inp_2,col_name = "name",bias_Col_name = "bias"):
    """
    Compares 2 GNSS clock bias DataFrames (from .clk), to a
    statistics table (with compar_clock_table)


    Parameters
    ----------
    DFclk_inp_1 & DFclk_inp_2 : DataFrame
        Clock DataFrame provided by files_rw.read_clk()

    Returns
    -------
    DFclk_diff : DataFrame
        Clock bias difference DataFrame
    """
    DF1idx = DFclk_inp_1.set_index([col_name,"epoch"])
    DF1idx.sort_index(inplace=True)
    
    DF2idx = DFclk_inp_2.set_index([col_name,"epoch"])
    DF2idx.sort_index(inplace=True)
    
    I1 = DF1idx.index
    I2 = DF2idx.index
    
    Iinter = I1.intersection(I2)
    Iinter = Iinter.sort_values()
    
    DF_diff_bias = DF1idx.loc[Iinter][bias_Col_name] - DF2idx.loc[Iinter][bias_Col_name]

    DFclk_diff = DF1idx.loc[Iinter].copy()
    DFclk_diff[bias_Col_name] = DF_diff_bias
    if "ac" in  DFclk_diff.columns:
        DFclk_diff.drop("ac",axis=1,inplace=True)
    else:
        DFclk_diff.drop("AC",axis=1,inplace=True)
    DFclk_diff.rename({bias_Col_name:bias_Col_name+"_diff"},inplace=True,axis=1)
    
    
    # Name definitions
    if 'ac' in DFclk_inp_1.columns:
        DFclk_diff['name1'] = DFclk_inp_1.ac.values[0]
    if 'AC' in DFclk_inp_1.columns:
        DFclk_diff['name1'] = DFclk_inp_1.AC.values[0]
    
    if 'ac' in DFclk_inp_1.columns:
        DFclk_diff['name2'] = DFclk_inp_2.ac.values[0]
    if 'AC' in DFclk_inp_1.columns:
        DFclk_diff['name2'] = DFclk_inp_2.AC.values[0]

    
    return DFclk_diff
    
def compar_clock_table(DFclk_diff_in,col_name = "name",bias_Col_name = "bias_diff"):
    """
    Generate a table with statistical indicators for a clock comparison
    (RMS mean, standard dev, ...)

    Parameters
    ----------
    DFclk_diff_in : DataFrame
        Clock bias difference DataFrame (from compar_clock)

    Returns
    -------
    DFcompar_out : DataFrame
        Statistical results of the comparison.

    """
    
    DF_diff_grp = DFclk_diff_in.groupby(col_name)[bias_Col_name]
    
    Smin  = DF_diff_grp.min().rename("min",inplace=True)
    Smax  = DF_diff_grp.max().rename("max",inplace=True)
    Smean = DF_diff_grp.mean().rename("mean",inplace=True)
    Sstd  = DF_diff_grp.std().rename("std",inplace=True)
    Srms  = DF_diff_grp.apply(stats.rms_mean).rename("rms",inplace=True)
    
    DFcompar_out = pd.concat([Smin,Smax,Smean,Sstd,Srms],axis=1)
    DFcompar_out.reset_index()
    
    return DFcompar_out

def compar_clk_plot(Diff_sat_all_df_in,
                    save_plot=False,
                    save_plot_dir="",
                    save_plot_name="auto",
                    save_plot_name_suffix=None,
                    save_plot_ext=(".pdf",".png",".svg"),
                    yaxis_limit=None,
                    yaxis_label_unit="psec",
                    col_name = 'name',
                    bias_Col_name = 'bias'):
    """
    General description

    Parameters
    ----------
    Diff_sat_all_df_in: DataFrame
        a DataFrame produced by compar_clk
        
    yaxis_limit: 3-tuple iterable or 2-element tuple
        force the y axis limits. must look like 
        (ymin,ymax) to set all th axis at the same limits 
    
    col_name: Normally the name of the column with the sat names
    bias_Col_name: The column with the clk values
    Default: 'bias'

    Returns
    -------
    the Figure and the 3 Axes if no save is asked
    export path (str) if save is asked
    but plot a plot anyway
    """
    
    
    import matplotlib.dates as mdates
    fig,axr = plt.subplots(1,1,sharex='all')
    Diff_sat_all_df_in = Diff_sat_all_df_in.reset_index()
    satdispo = natsort.natsorted(list(set(Diff_sat_all_df_in[col_name])))
    # satdispo = natsort.natsorted(list(set(Diff_sat_all_df_in['sat'])))
       
    SymbStk = []
       
    cm = plt.get_cmap('viridis')
    NUM_COLORS = len(satdispo)
    Colors = [cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)]
       
    
    
    Date = conv.numpy_dt2dt(Diff_sat_all_df_in.epoch.values[0])
    Diff_sat_all_df_in.name = ' '.join(('Clock comparison  b/w',
                                  Diff_sat_all_df_in.name1.values[0] ,'(ref.) and',
                                  Diff_sat_all_df_in.name2.values[0] ,',',Date.strftime("%Y-%m-%d")))
    # Pandas donesn't manage well iterable as attribute
    # So, it is separated
    try:
        col_name0 = Diff_sat_all_df_in.frame_col_name1
        col_name1 = Diff_sat_all_df_in.frame_col_name2
        col_name2 = Diff_sat_all_df_in.frame_col_name3
    except:
        col_name0 = Diff_sat_all_df_in.columns[0]
        col_name1 = Diff_sat_all_df_in.columns[1]
        col_name2 = Diff_sat_all_df_in.columns[2]
       
    for satuse,color in zip(satdispo,Colors):
        Diffuse = Diff_sat_all_df_in[Diff_sat_all_df_in[col_name] == satuse]
       
        Time = Diffuse.epoch
        R    = Diffuse[bias_Col_name+'_diff']*10**12
       
       
        #fig.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')
       
        Symb = axr.plot(Time,R,label=satuse,c=color)
       
       
        SymbStk.append(Symb[0])
       
        fig.autofmt_xdate()
       
       
    ylabuni = " (" + yaxis_label_unit + ")"
    
       
    axr.set_ylabel('Bias Diff.' + ylabuni)
       
       
       
    y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)
    axr.yaxis.set_major_formatter(y_formatter)
       
    
    if yaxis_limit and len(yaxis_limit) == 3: ### indep. axis limit
        axr.set_ylim(yaxis_limit[0])
       
    elif yaxis_limit and len(yaxis_limit) == 2:
        axr.set_ylim(yaxis_limit)
       
    else:
        pass
        
        
    import matplotlib.dates as mdates
    fig.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')
       
    lgd = fig.legend(tuple(SymbStk), satdispo , loc='lower center',ncol=8,
                     columnspacing=1)
       
    fig.set_size_inches(8.27,11.69)
    plt.suptitle(Diff_sat_all_df_in.name)
    plt.tight_layout()
    plt.subplots_adjust(top=0.95)
    plt.subplots_adjust(bottom=0.15)
       
    if save_plot:
        if save_plot_name == "auto":
            save_plot_name = Diff_sat_all_df_in.name1.values[0]+"_"+Diff_sat_all_df_in.name2.values[0]+"_"+Date.strftime("%Y-%m-%d")
            
        if save_plot_name_suffix:
            save_plot_name = save_plot_name + '_' + save_plot_name_suffix
       
        for ext in save_plot_ext:
            save_plot_path = os.path.join(save_plot_dir,save_plot_name)
            plt.savefig(save_plot_path + ext)
            return_val = save_plot_path
            
    else:
        return_val = fig,(axr)

    return return_val






def compar_sinex(snx1 , snx2 , stat_select = None, invert_select=False,
                 out_means_summary=True,out_meta=True,out_dataframe = True,
                 manu_wwwwd=None):

    if type(snx1) is str:
        week1 = utils.split_improved(os.path.basename(snx1),"_",".")[:]
        week2 = utils.split_improved(os.path.basename(snx2),"_",".")[:]
        if week1 != week2:
            log.warning("Dates of 2 input files are differents !!! It might be very bad !!! %s %s",week1,week2)
        else:
            wwwwd = week1
        D1 = files_rw.read_sinex(snx1,True)
        D2 = files_rw.read_sinex(snx2,True)
    else:
        log.warning("WARN : you are giving the SINEX input as a DataFrame, wwwwd has to be given manually using manu_wwwwd")
        D1 = snx1
        D2 = snx2


    if manu_wwwwd:
        wwwwd = manu_wwwwd


    STATCommon  = set(D1["STAT"]).intersection(set(D2["STAT"]))

    if stat_select:

        STATCommon_init = list(STATCommon)

        if invert_select:
            select_fct = lambda x : not x
        else:
            select_fct = lambda x : x

        if type(stat_select) is str:
            STATCommon = [sta for sta in STATCommon_init if select_fct(re.search(stat_select, sta)) ]
        elif utils.is_iterable(stat_select):
            STATCommon = [sta for sta in STATCommon_init if select_fct(sta in stat_select) ]
        else:
            log.warning("WARN : check type of stat_select")

    D1Common = D1[D1["STAT"].isin(STATCommon)].sort_values("STAT").reset_index(drop=True)
    D2Common = D2[D2["STAT"].isin(STATCommon)].sort_values("STAT").reset_index(drop=True)


    Ddiff = pd.DataFrame()
    Ddiff = Ddiff.assign(STAT=D1Common["STAT"])

    #### XYZ Part
    for xyz in ("x","y","z"):

        dif = pd.to_numeric((D2Common[xyz] - D1Common[xyz]))

        Ddiff = Ddiff.assign(xyz=dif)
        Ddiff = Ddiff.rename(columns={"xyz": xyz})

    D3D = np.sqrt((Ddiff["x"]**2 + Ddiff["y"]**2 + Ddiff["z"]**2 ).astype('float64'))

    Ddiff = Ddiff.assign(d3D_xyz=D3D)

    ### ENU Part
    E , N ,U = [] , [] , []
    enu_stk = []

    for (_,l1) , (_,l2) in zip( D1Common.iterrows() , D2Common.iterrows() ):
        enu   = conv.XYZ2ENU_2(l1["x"],l1["y"],l1["z"],l2["x"],l2["y"],l2["z"])
        enu_stk.append(np.array(enu))


    if len(enu_stk) == 0:
        E,N,U = np.array([]) , np.array([]) , np.array([])
    else:
        ENU = np.hstack(enu_stk)
        E,N,U = ENU[0,:] , ENU[1,:] , ENU[2,:]


    D2D = np.sqrt((E**2 + N**2).astype('float64'))
    D3D = np.sqrt((E**2 + N**2 + U**2 ).astype('float64'))

    Ddiff = Ddiff.assign(e=E)
    Ddiff = Ddiff.assign(n=N)
    Ddiff = Ddiff.assign(u=U)
    Ddiff = Ddiff.assign(d2D_enu=D2D)
    Ddiff = Ddiff.assign(d3D_enu=D3D)

    #    E,N,U    = conv.XYZ2ENU_2((X,Y,Z,x0,y0,z0))
    #    E,N,U    = conv.XYZ2ENU_2((X,Y,Z,x0,y0,z0))

    if out_dataframe:
        out_meta = True


    if not out_means_summary:
        log.info("INFO : this is not used operationally and it can be improved")
        return Ddiff
    else:
        output = []

        col_names = ("x","y","z","d3D_xyz",
                     "e","n","u","d2D_enu","d3D_enu")

        for xyz in col_names:
            output.append(stats.rms_mean(Ddiff[xyz]))
        for xyz in col_names:
            output.append(np.nanmean(Ddiff[xyz]))
        for xyz in col_names:
            output.append(np.nanstd(Ddiff[xyz]))

        if out_meta:
            nstat = len(STATCommon)
            week   = int(wwwwd[:4])
            day    = int(wwwwd[4:])
            output = [week , day ,nstat] + output


        if not out_dataframe:
            return tuple(output)
        else:

            output_DF = pd.DataFrame(output).transpose()

            output_DF.columns = ["week","dow","nbstat",
             "x_rms","y_rms","z_rms","d3D_xyz_rms",
             "e_rms","n_rms","u_rms","d2D_enu_rms","d3D_enu_rms",
             "x_ari","y_ari","z_ari","d3D_xyz_ari",
             "e_ari","n_ari","u_ari","d2D_enu_ari","d3D_enu_ari",
             "x_ari","y_std","z_std","d3D_xyz_std",
             "e_ari","n_std","u_std","d2D_enu_std","d3D_enu_std"]

            return output_DF


#   ____       _     _ _     _____        _        ______                             
#  / __ \     | |   (_) |   |  __ \      | |      |  ____|                            
# | |  | |_ __| |__  _| |_  | |  | | __ _| |_ __ _| |__ _ __ __ _ _ __ ___   ___  ___ 
# | |  | | '__| '_ \| | __| | |  | |/ _` | __/ _` |  __| '__/ _` | '_ ` _ \ / _ \/ __|
# | |__| | |  | |_) | | |_  | |__| | (_| | || (_| | |  | | | (_| | | | | | |  __/\__ \
#  \____/|_|  |_.__/|_|\__| |_____/ \__,_|\__\__,_|_|  |_|  \__,_|_| |_| |_|\___||___/
#                                                                                     
       
### Orbit DataFrames   

def OrbDF_lagrange_interpolate(DForb_in,Titrp,n=10,
                               append_to_input_DF = False,
                               plot=False):
    """
    High level function to interpolate an orbit DataFrame

    Parameters
    ----------
    DForb_in : DataFrame
        an Orbit DataFrame.
    Titrp : iterable of datetime
        Epochs of the wished points.
    n : int, optional
        degree of the polynom. Better if even. The default is 10.
    append_to_input_DF : bool, optional
        append the interpolated DF to the input DF. The default is False.
    plot : bool, optional
        Plot the values. For debug only. The default is False.

    Returns
    -------
    DForb_out : DataFrame
        Interpolated orbits.
        
    Tips
    ----
    Use conv.dt_range to generate the wished epochs range

    """
    DForb_stk = []
    
    for sat,ac in itertools.product(DForb_in.sat.unique(),DForb_in['ac'].unique()):
        
        log.info("process",ac,sat)
        
        DForb_use = DForb_in[(DForb_in.sat == sat) & (DForb_in['ac'] == ac)].copy()
            
        Tdata = DForb_use.epoch.dt.to_pydatetime()
        
        Xitrp = stats.lagrange_interpolate(Tdata,DForb_use.x,Titrp,n=n)
        Yitrp = stats.lagrange_interpolate(Tdata,DForb_use.y,Titrp,n=n)
        Zitrp = stats.lagrange_interpolate(Tdata,DForb_use.z,Titrp,n=n)
        
        ClkDummy = np.array([999999.999999] * len(Titrp))
        
        ARR = np.column_stack((Titrp,Xitrp,Yitrp,Zitrp,ClkDummy))
        
        DForb_tmp = pd.DataFrame(ARR,columns=["epoch","x","y","z","clk"])
        
        ### sometihng else must be tested o give the annex val directly in the col of DForb_tmp
        DFannex_vals = DForb_use.drop(["epoch","x","y","z","clk"],axis=1).drop_duplicates()
        DFannex_vals = pd.concat([DFannex_vals]*(len(Titrp)),ignore_index=True,axis=0)
        
        DForb_tmp = pd.concat((DForb_tmp,DFannex_vals),axis=1)
        DForb_stk.append(DForb_tmp)
        
        if plot:
            # plt.plot(Tdata,DForb_use.x,'o')
            # plt.plot(Titrp,Xitrp,'.')  
            ## GUS mod 220322
            fig,axr = plt.subplots(1,1,sharex='all')
            Symb = axr.plot(Tdata,DForb_use.x,'o')
            Symb = axr.plot(Titrp,Xitrp,'.')

        
    
    DForb_out = pd.concat(DForb_stk)
    
    if append_to_input_DF:
        DForb_out = pd.concat((DForb_in,DForb_out))
        
    DForb_out.reset_index(drop=True)
    DForb_out[["x","y","z","clk"]] = DForb_out[["x","y","z","clk"]].astype(float)
    return DForb_out

def OrbDF_crf2trf(DForb_inp,DF_EOP_inp,time_scale_inp="gps",
                  inv_trf2crf=False):
    """
    Convert an Orbit DataFrame from Celetrial Reference Frame to 
    Terrestrial Reference Frame (.
    
    Requires EOP to work. Cf. note below.

    Parameters
    ----------
    DForb_inp : DataFrame
        Input Orbit DataFrame in Celetrial Reference Frame.
    DF_EOP_inp : DataFrame
        EOP DataFrame  (C04 format).
    time_scale_inp : str, optional
        The time scale used in. manage 'utc', 'tai' and 'gps'.
        The default is "gps".
    inv_trf2crf : bool, optional
        Provide the inverse transformation TRF => CRF.
        The default is False.

    Returns
    -------
    DForb_out : DataFrame
        Output Orbit DataFrame in Terrestrial Reference Frame.
        (or Celestrial if inv_trf2crf is True)
        
    Note
    ----
    The EOP can be obtained from the IERS C04 products.
    e.g.
    https://datacenter.iers.org/data/latestVersion/224_EOP_C04_14.62-NOW.IAU2000A224.txt
    To get them as a Compatible DataFrame, use the function
    files_rw.read_eop_C04()
    """
    
    DForb = DForb_inp.copy()
    
    import geodezyx.reffram.sofa18 as sofa
    
    ### bring everything to UTC
    if time_scale_inp.lower() == "gps":
        DForb["epoch_utc"] = conv.dt_gpstime2dt_utc(DForb["epoch"])
    elif time_scale_inp.lower() == "tai":
        DForb["epoch_utc"] = conv.dt_tai2dt_utc(DForb["epoch"])
    elif time_scale_inp.lower() == "utc":
        DForb["epoch_utc"] = DForb["epoch"]
    ### TT and UT1 are not implemented (quite unlikely to have them as input)
    
    ### do the time scale's conversion
    DForb["epoch_tai"] = conv.dt_utc2dt_tai(DForb["epoch_utc"])
    DForb["epoch_tt"]  = conv.dt_tai2dt_tt(DForb["epoch_tai"])
    DForb["epoch_ut1"] = conv.dt_utc2dt_ut1_smart(DForb["epoch_utc"],DF_EOP_inp)
        
    ### Do the EOP interpolation 
    DF_EOP_intrp = eop_interpotate(DF_EOP_inp, DForb["epoch_utc"])
    ### bring the EOP to radians
    Xeop = np.deg2rad(conv.arcsec2deg(DF_EOP_intrp['x']))
    Yeop = np.deg2rad(conv.arcsec2deg(DF_EOP_intrp['y']))
    
    TRFstk = []
    
    for tt,ut1,xeop,yeop,x,y,z in zip(DForb["epoch_tt"],
                                      DForb["epoch_ut1"],
                                      Xeop,Yeop,
                                      DForb['x'],DForb['y'],DForb['z']):
    
        MatCRF22TRF = sofa.iau_c2t06a(2400000.5,
                                      conv.dt2MJD(tt),
                                      2400000.5,
                                      conv.dt2MJD(ut1),
                                      xeop,yeop)
        if inv_trf2crf:
            MatCRF22TRF = np.linalg.inv(MatCRF22TRF)
    
        CRF = np.array([x,y,z])
        TRF = np.dot(MatCRF22TRF,CRF)
    
        TRFstk.append(TRF)
    
    ### Final stack and replacement
    TRFall = np.vstack(TRFstk)
    DForb_out = DForb_inp.copy()
    DForb_out[["x","y","z"]] = TRFall
    
    return DForb_out
                                                                     

#### FCT DEF
def OrbDF_reg_2_multidx(OrbDFin,index_order=["prn","epoch"]):
    """
    From an regular Orbit DF generated by read_sp3(), set some columns 
    (typically ["sat","epoch"]) as indexes
    The outputed DF is then a Multi-index DF
    """
    OrbDFwrk = OrbDFin.reset_index()
    OrbDFwrk = OrbDFwrk.sort_values(index_order)
    OrbDFwrk = OrbDFwrk.set_index(index_order,inplace=False)
    return OrbDFwrk

def OrbDF_multidx_2_reg(OrbDFin,index_order=["sat","epoch"]):
    """
    Convert a Multi-index formatted OrbitDF to his original form
    """
    OrbDFwrk = OrbDFin.reset_index()
    OrbDFwrk = OrbDFwrk.sort_values(index_order)
    OrbDFwrk["sys"] = OrbDFwrk["prn"].apply(lambda x: x[0])
    OrbDFwrk["prni"] = OrbDFwrk["prn"].apply(lambda x: int(x[1:]))
    return OrbDFwrk

def OrbDF_common_epoch_finder(OrbDFa_in,OrbDFb_in,return_index=False,
                              supplementary_sort=False,order=["prn","epoch"]):
    """
    Find common sats and epochs in to Orbit DF,
    and output the corresponding Orbit DFs
    order >> normally for sp3 is sat and epoch, 
    but can be used for snx files as STAT and epoch
    """
    
    OrbDFa = OrbDF_reg_2_multidx(OrbDFa_in,index_order = order)
    OrbDFb = OrbDF_reg_2_multidx(OrbDFb_in,index_order = order)
    
    I1 = OrbDFa.index
    I2 = OrbDFb.index
    
    Iinter = I1.intersection(I2)
    ### A sort of the Index to avoid issues ...
    Iinter = Iinter.sort_values()
    
    OrbDFa_out = OrbDFa.loc[Iinter]
    OrbDFb_out = OrbDFb.loc[Iinter]
    
    if supplementary_sort:
        # for multi GNSS, OrbDF_out are not well sorted (why ??? ...)
        # we do a supplementary sort
        # NB 202003: maybe because Iiter was not sorted ...
        # should be fixed with the Iinter.sort_values() above 
        # but we maintain this sort
        OrbDFa_out = OrbDFa_out.sort_values(order)
        OrbDFb_out = OrbDFb_out.sort_values(order)

    
    if len(OrbDFa_out) != len(OrbDFb_out):
        log.warning("len(Orb/ClkDFa_out) != len(Orb/ClkDFb_out)")
        log.warning("TIPS : ClkDFa_in and/or ClkDFb_in might contain duplicates")
    
    if return_index:
        return OrbDFa_out , OrbDFb_out , Iinter
    else:
        return OrbDFa_out , OrbDFb_out


def OrbDF_const_sv_columns_maker(OrbDFin,inplace=True):
    """
    (re)generate the const and sv columns from the sat one
    """
    if inplace:
        OrbDFin['sys'] = OrbDFin['prn'].str[0]
        OrbDFin['prni']    = OrbDFin['prn'].apply(lambda x: int(x[1:]))
        return None
    else:
        OrbDFout = OrbDFin.copy()
        OrbDFout['sys'] = OrbDFout['prn'].str[0]
        OrbDFout['prni']    = OrbDFout['prn'].apply(lambda x: int(x[1:]))
        return OrbDFout

 #   _____ _            _      _____        _        ______                              
 #  / ____| |          | |    |  __ \      | |      |  ____|                             
 # | |    | | ___   ___| | __ | |  | | __ _| |_ __ _| |__ _ __ __ _ _ __ ___   ___  ___  
 # | |    | |/ _ \ / __| |/ / | |  | |/ _` | __/ _` |  __| '__/ _` | '_ ` _ \ / _ \/ __| 
 # | |____| | (_) | (__|   <  | |__| | (_| | || (_| | |  | | | (_| | | | | | |  __/\__ \ 
 #  \_____|_|\___/ \___|_|\_\ |_____/ \__,_|\__\__,_|_|  |_|  \__,_|_| |_| |_|\___||___/ 
                                                                                                                                                                        
### Clock DataFrames   

def ClkDF_filter(ClkDF_in,
             typ=("AS","AR"),
             name=None,
             ac=None,
             epoch_strt=dt.datetime(1980,1,1),
             epoch_end=dt.datetime(2099,1,1),
             name_regex=False):
    """
    Filter the content of a Clock DataFrame

    Parameters
    ----------
    ClkDF_in : DataFrame
        Input Clock DataFrame 
        (a concatenation of DF generated by files_rw.read_clk.
    typ : iterable of str, optional
        List of the types of clocks: AS (satellite) or AR (receiver).
        The default is ("AS","AR").
    name : iterable of str, optional
        List of wished satellites/stations.
        Can be a regex (see also name_regex)
        The default is None.
    ac : iterable of str, optional
        List of wished ACs. The default is None.
    epoch_strt : datetime, optional
        Start epoch. The default is dt.datetime(1980,1,1).
    epoch_end : datetime, optional
        End epoch (not included). The default is dt.datetime(2099,1,1).
    name_regex : bool, optional
        the given names as 'name' arguments are regular expressions
        Some useful regex are given bellow
        The default is False

    Returns
    -------
    Clock DataFrame
        Output Clock DataFrame.
        
    Notes
    -----
    '^E[0-9]{2}': Galileo Satellites
    '^G[0-9]{2}': GPS Satellites

    """
    
    if type(ClkDF_in) is str:
        ClkDF_wrk = utils.pickle_loader(ClkDF_in)
    else:
        ClkDF_wrk = ClkDF_in
    
    BOOL = np.ones(len(ClkDF_wrk)).astype(bool)
    
    if typ:
        BOOLtmp = ClkDF_wrk.type.isin(typ)
        BOOL    = BOOL & np.array(BOOLtmp)

    if name:
        if not name_regex: ### full name mode
            BOOLtmp = ClkDF_wrk.name.isin(name)
            BOOL    = BOOL & np.array(BOOLtmp)   
        else: ### REGEX mode
            BOOLtmp = np.zeros(len(ClkDF_wrk.name)).astype(bool)
            for rgx in name:
                NamSerie = ClkDF_wrk.name
                BOOLtmp = BOOLtmp | np.array(NamSerie.str.contains(rgx))

            BOOL = BOOL & np.array(BOOLtmp)
                 
    if ac:
        BOOLtmp = ClkDF_wrk.ac.isin(ac)
        BOOL    = BOOL & np.array(BOOLtmp)    
        
    ##epoch
    BOOLtmp = (epoch_strt <= ClkDF_wrk.epoch) & (ClkDF_wrk.epoch < epoch_end)
    BOOL    = BOOL & np.array(BOOLtmp)    
    
    return ClkDF_wrk[BOOL]

def ClkDF_reg_2_multidx(ClkDFin,index_order=["name","epoch"]):
    """
    From an regular Clock DF generated by read_clk(), set some columns 
    (typically ["name","epoch"]) as indexes
    The outputed DF is then a Multi-index DF
    
    It an adapted version of OrbDF_reg_2_multidx
    """
    
    return OrbDF_reg_2_multidx(ClkDFin,index_order)
    

def ClkDF_common_epoch_finder(ClkDFa_in,ClkDFb_in,return_index=False,
                              supplementary_sort=False,
                              order=["name","epoch"]):
    """
    Find common sats/station and epochs in to Clock DF, and output the
    corresponding Clock DFs
    
    Is an adapted version of OrbDF_common_epoch_finder
    """
    
    return OrbDF_common_epoch_finder(ClkDFa_in,ClkDFb_in,
                                     return_index=return_index,
                                     supplementary_sort=supplementary_sort,
                                     order=order)



def ClkDF_common_epoch_finder_multi(ClkDF_list_in,
                                    return_index=False,
                                    supplementary_sort=False,
                                    order=["name","epoch"]):
    
    """
    Find common sats/station and epochs in to Clock DF, and output the
    corresponding Clock DFs
    
    Is is the multi version of ClkDF_common_epoch_finder
    """
    
    ClkDFref = ClkDF_list_in[0]
    
    #### First loop: we find the common epochs
    for ClkDF in ClkDF_list_in[1:]:
        
        OUTTUP = OrbDF_common_epoch_finder(ClkDFref,ClkDF,
                                           return_index=True,
                                           supplementary_sort=supplementary_sort,
                                           order=order)
        
        ClkDFref , _ , Iinter = OUTTUP
        
    #### second loop: we use the common epochs found for the outputed ClkDF  
    ClkDF_list_out= []
    for ClkDF in ClkDF_list_in:
        ClkDFout = ClkDF.set_index(order).loc[Iinter]
        ClkDF_list_out.append(ClkDFout)
    
    if not return_index:
        return ClkDF_list_out
    else:
        return ClkDF_list_out,Iinter
        


 #   _____ _      _____   __      __   _ _     _       _   _             
 #  / ____| |    |  __ \  \ \    / /  | (_)   | |     | | (_)            
 # | (___ | |    | |__) |  \ \  / /_ _| |_  __| | __ _| |_ _  ___  _ __  
 #  \___ \| |    |  _  /    \ \/ / _` | | |/ _` |/ _` | __| |/ _ \| '_ \ 
 #  ____) | |____| | \ \     \  / (_| | | | (_| | (_| | |_| | (_) | | | |
 # |_____/|______|_|  \_\     \/ \__,_|_|_|\__,_|\__,_|\__|_|\___/|_| |_|
                                                                       


def svn_prn_equiv_DF(path_meta_snx):
    """
    generate a SVN <> PRN equivalent DataFrame

    Parameters
    ----------
    path_meta_snx : str
        path of the MGEX metadata sinex.
        last version avaiable here
        http://mgex.igs.org/IGS_MGEX_Metadata.php

    Returns
    -------
    DFfin : Pandas DataFrame
        SVN <> PRN equivalent DataFrame.

    """
    
    DFsvn  = files_rw.read_sinex_versatile(path_meta_snx,"SATELLITE/IDENTIFIER",
                                           header_line_idx=-2)    

    DFprn  = files_rw.read_sinex_versatile(path_meta_snx,"SATELLITE/PRN",
                                           header_line_idx=-2)
    
    DFsvn.drop(columns='Comment__________________________________',inplace=True)
    DFprn.drop(columns='Comment_________________________________',inplace=True)

    ## the next lines 1 and 3 seems like they have became useless
    DFsvn["SVN_"] = DFsvn["SVN_"].apply(lambda x:x[0] + x[1:])
    DFprn.replace(dt.datetime(1970,1,1),dt.datetime(2099,1,1),inplace=True)
    DFprn["SVN_"] = DFprn["SVN_"].apply(lambda x:x[0] + x[1:])
    
    
    DFstk = []
    
    for isat , sat in DFprn.iterrows():
        svn = sat["SVN_"]
        
        sat["Block"] = DFsvn[DFsvn["SVN_"] == svn]["Block__________"].values[0]
        DFstk.append(sat)
        
    DFfin = pd.concat(DFstk,axis=1).transpose()
    
    DFfin.rename(columns={"SVN_":"SVN",
                          "Valid_From____":"start",
                          "Valid_To______":"end"},inplace=True)
    
    
    DFfin["const"]   = DFfin["SVN"].apply(lambda x:x[0])
    DFfin["SVN_int"] = DFfin["SVN"].apply(lambda x:int(x[1:]))
    DFfin["PRN_int"] = DFfin["PRN"].apply(lambda x:int(x[1:]))    
    
    return DFfin
    

def svn_prn_equiv(sat_in,date_in,
                  svn_prn_equiv_DF,
                  mode="svn2prn",
                  full_output=False):
    """
    Get the equivalence SVN <> PRN for a given epoch
    
    Parameters
    ----------
    sat_in : str
        Satellite "ID", SVN or PRN.
    date_in : datetime
        wished epoch.
    svn_prn_equiv_DF : DataFrame
        Equivalence table generated by svn_prn_equiv_DF.
    mode : str, optional
        prn2svn: PRN > SVN
        svn2prn: SVN > PRN.
        The default is "svn2prn".
    full_output : bool, optional
        get the complete Equivalence table row. The default is False.

    Returns
    -------
    str or DataFrame
    """

    svnorprn1 = mode[:3].upper()    
    svnorprn2 = mode[-3:].upper()
    
    DFsat = svn_prn_equiv_DF[svn_prn_equiv_DF[svnorprn1] == sat_in]
    Bool_date = np.logical_and((DFsat.start <= date_in) , (date_in < DFsat.end))
    DFout = DFsat[Bool_date]
    
    if len(DFout) != 1:
        log.warning("several or no %s entries !!! %s %s",mode,sat_in,date_in)
        
    if full_output:
        return DFout
    else:
        return DFout[svnorprn2].values[0]
    
    
def get_block_svn(sat_in,svn_prn_equiv_DF):
    """
    Get the equivalence SVN block type
    
    Parameters
    ----------
    sat_in : str
        Satellite SVN.
    svn_prn_equiv_DF : DataFrame
        Equivalence table generated by svn_prn_equiv_DF.
    Returns
    -------
    str with the block name
    """
    DFsat = svn_prn_equiv_DF[svn_prn_equiv_DF['SVN'] == sat_in]
    # if DFsat.empty:
    #     print('SVN NOT FOUND')
    # else:        
    block = DFsat.Block.values[0]

    return block    


def stats_slr(DFin,grpby_keys = ['sat'],
              threshold = .5):
    """
    computes statistics for SLR Residuals
    
    Parameters
    ----------
    DFin : Pandas DataFrame
        Input residual Dataframe from read_pdm_res_slr.
    grpby_keys : list of str, optional
        The default is ['sat'].
        per day, per solution, per satellite: ['day','sol','sat']
        per day, per solution, per station: ['day','sol','sta']
        per day, per solution, per satellite, per station: ['day','sol','sta','sat']
    threshold : float
        apply a Threshold

    Returns
    -------
    DD : Output statistics DataFrame
        return the mean, the rms and the std.
    """
    
    DD = DFin[np.abs(DFin["res"]) < threshold]
    
    DD_grp  = DD.groupby(grpby_keys)
    DD_mean = DD_grp['res'].agg(np.mean).rename('mean') * 1000
    DD_rms  = DD_grp['res'].agg(stats.rms_mean).rename('rms')   * 1000
    DD_std  = DD_grp['res'].agg(np.std).rename('std')   * 1000
    DD = pd.concat([DD_mean,DD_std,DD_rms],axis=1)
    DD.reset_index(inplace = True)
    
    return DD    


 #  ______           _   _        ____       _            _        _   _               _____                               _                
 # |  ____|         | | | |      / __ \     (_)          | |      | | (_)             |  __ \                             | |               
 # | |__   __ _ _ __| |_| |__   | |  | |_ __ _  ___ _ __ | |_ __ _| |_ _  ___  _ __   | |__) |_ _ _ __ __ _ _ __ ___   ___| |_ ___ _ __ ___ 
 # |  __| / _` | '__| __| '_ \  | |  | | '__| |/ _ \ '_ \| __/ _` | __| |/ _ \| '_ \  |  ___/ _` | '__/ _` | '_ ` _ \ / _ \ __/ _ \ '__/ __|
 # | |___| (_| | |  | |_| | | | | |__| | |  | |  __/ | | | || (_| | |_| | (_) | | | | | |  | (_| | | | (_| | | | | | |  __/ ||  __/ |  \__ \
 # |______\__,_|_|   \__|_| |_|  \____/|_|  |_|\___|_| |_|\__\__,_|\__|_|\___/|_| |_| |_|   \__,_|_|  \__,_|_| |_| |_|\___|\__\___|_|  |___/
                                                                                                                                         

### EOP / Earth Oreintation Parameters

def eop_interpotate(DF_EOP,Epochs_intrp,eop_params = ["x","y"]):
    """
    Interopolate the EOP provided in a C04-like DataFrame

    Parameters
    ----------
    DF_EOP : DataFrame
        Input EOP DataFrame (C04 format).
        Can be generated by files_rw.read_eop_C04
    Epochs_intrp : datetime of list of datetimes
        Wished epochs for the interpolation.
    eop_params : list of str, optional
        Wished EOP parameter to be interpolated.
        The default is ["x","y"].

    Returns
    -------
    OUT : DataFrame or Series
        Interpolated parameters.
        Series if onely one epoch is provided, DF_EOP elsewere
    """
    if not utils.is_iterable(Epochs_intrp):
        singleton = True
    else:
        singleton = False
    
    I_eop   = dict()
    Out_eop = dict()
    Out_eop["epoch"] = Epochs_intrp    
    
    for eoppar in eop_params:
        I = conv.interp1d_time(DF_EOP.epoch,DF_EOP[eoppar])
        I_eop[eoppar] = I
        try:
            Out_eop[eoppar] = I(Epochs_intrp)
        except ValueError as err:
            log.error("in EOP interpolation")
            log.error("param.: %s, epoch: %s",eoppar,Epochs_intrp)
            raise err
      
    if not singleton:
        OUT = pd.DataFrame(Out_eop)
    else:
        OUT = pd.Series(Out_eop)
        
    return OUT
